from math import ceil
from typing import List

from scipy.io import wavfile

from py_headless_daw.compiler.project_compiler import ProjectCompiler
from py_headless_daw.project.project import Project
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.wiring import StreamNode

import numpy as np


class ProjectRenderer:
    SAMPLE_RATE = 44100

    def __init__(self, project_compiler: ProjectCompiler):
        self._compiler: ProjectCompiler = project_compiler

    def render_to_file(self, project: Project, start_time: float, end_time: float, out_file_path: str):
        buffer = self.render_to_array(project, start_time, end_time)
        buffer = np.rot90(buffer, k=-1)
        wavfile.write(out_file_path, self.SAMPLE_RATE, buffer)

    def render_to_array(self, project: Project, start_time: float, end_time: float) -> np.ndarray:
        output_stream_nodes: List[StreamNode] = self._compiler.compile(project)

        left_node = output_stream_nodes[0]
        right_node = output_stream_nodes[1]

        buffer_length_samples = 4410
        sample_rate = self.SAMPLE_RATE

        buffer_length_seconds = buffer_length_samples / sample_rate
        num_frames = ceil((end_time - start_time) / buffer_length_seconds)

        result_buffer = np.ndarray(shape=(2, num_frames * buffer_length_samples), dtype=np.float32)

        left_frame_buffer = np.ndarray(shape=(buffer_length_samples,), dtype=np.float32)
        right_frame_buffer = np.ndarray(shape=(buffer_length_samples,), dtype=np.float32)

        for i in range(num_frames):
            interval = TimeInterval()

            interval.num_samples = buffer_length_samples
            interval.start_in_seconds = start_time + i * buffer_length_seconds
            interval.end_in_seconds = interval.start_in_seconds + buffer_length_seconds

            left_node.render(interval, left_frame_buffer)
            right_node.render(interval, right_frame_buffer)

            start_sample_in_result = i * buffer_length_samples
            end_sample_in_result = start_sample_in_result + buffer_length_samples

            result_buffer[0][start_sample_in_result:end_sample_in_result] = left_frame_buffer
            result_buffer[1][start_sample_in_result:end_sample_in_result] = right_frame_buffer

        return result_buffer



import os
import unittest
from typing import List

import numpy as np

from py_headless_daw.processing.hybrid.vst_plugin import VstPlugin
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.midi_event_factory import MidiEventFactory
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit


class VstPluginProcessingStrategyTest(unittest.TestCase):

    def test_vst_plugin_strategy(self):
        host = Host()

        dir_path = os.path.dirname(os.path.realpath(__file__))

        path_to_so: str = dir_path + '/../test_plugins/amsynth-vst.x86_64-linux.so'
        as_bytes = path_to_so.encode('utf-8')

        unit = Unit(0, 0, 1, 0, host)
        strategy = VstPlugin(as_bytes, unit)

        unit.set_processing_strategy(strategy)
        event_factory = MidiEventFactory()
        event = event_factory.create_note_on_event(87, 10)
        event.sample_position = 9
        events: List[Event] = [event]

        def generate():
            return np.zeros((host.block_size,), np.float32)

        left_out = generate()
        right_out = generate()

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1
        interval.num_samples = 100

        self.assertTrue(left_out[13] == 0.0)
        self.assertTrue(right_out[13] == 0.0)

        strategy.render(interval, [], [left_out, right_out], [events], [])

        self.assertTrue(left_out[13] != 0.0)
        self.assertTrue(right_out[13] != 0.0)



from py_headless_daw.project.having_parameters import HavingParameters


class NamedParameterBag(HavingParameters):
    def __init__(self):
        super().__init__()
        self.name: str = 'unnamed'



from py_headless_daw.project.content.midi_clip import MidiClipEvent, MidiClip


class MidiNote(MidiClipEvent):

    def __init__(self, clip, clip_time, note, velocity, length):
        # type: (MidiClip, float, int, int, float)->None
        super().__init__(clip, clip_time)
        self.note: int = note
        self.velocity: int = velocity
        self.length: float = length

    def get_absolute_end_time(self) -> float:
        return self.get_absolute_start_time() + self.length



import unittest
from copy import copy

import numpy as np

from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator import DrumSynthGenerator
from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator_config import DrumSynthGeneratorConfig, OscillatorConfig
from py_headless_daw.dsp_utils.wave_types import WAVE_TYPE_SINE


# import matplotlib.pyplot as plt


class MyTestCase(unittest.TestCase):
    def test_generate_bd_with_two_oscillators(self):
        osc_config1 = OscillatorConfig(
            wave_type=WAVE_TYPE_SINE,
            volume=1,
            zero_frequency=0,
            frequency_range=300,
            start_phase=0,

            volume_envelope_attack_time=0.001,

            volume_envelope_decay_time=0.1,
            volume_envelope_decay_curve_power=4,
            volume_envelope_decay_curve_ratio=-1,

            volume_envelope_sustain_level=0,

            pitch_envelope_attack_time=0,
            pitch_envelope_decay_time=0.1,
            pitch_envelope_decay_curve_ratio=-0.9,
            pitch_envelope_decay_curve_power=4,
            pitch_envelope_sustain_level=0
        )

        osc_config2 = copy(osc_config1)
        osc_config2.start_phase = 2.123
        osc_config2.frequency_range = 200

        config = DrumSynthGeneratorConfig([osc_config1, osc_config2])
        generator = DrumSynthGenerator(config)

        output_buffer_size = 5000
        output_buffer = np.ndarray(shape=(output_buffer_size,), dtype=np.float32)

        generator.render_to_buffer(output_buffer, 44100, 0)

        # plt.plot(output_buffer)
        # plt.show()


if __name__ == '__main__':
    unittest.main()



from typing import Sequence

from py_headless_daw.project.content.clip import Clip


class MidiClip(Clip):
    def __init__(self, start_time: float, end_time: float):
        super().__init__(start_time, end_time)
        from py_headless_daw.project.content.midi_note import MidiNote  # to avoid an import-time loop
        self.midi_notes: Sequence[MidiNote] = []

    @property
    def events(self):
        # type: ()->Sequence[MidiClipEvent]
        return self.midi_notes


class MidiClipEvent:
    """
    A class representing a midi event

    A midi clip event is an instant (zero length) something that is characterized by the time of its onset.
    There are two ways to translate clip times to wall clock times:

    TIMING_TYPE_CLIP_RELATIVE - the time offset from the beginning of the clips is calculated
    proportionally to the length of the clip

    TIMING_TYPE_CLIP_ABSOLUTE - the time offset from the beginning of the clips is
    taken directly as number of seconds (float) to pass from the beginning of the clip.
    """
    TIMING_TYPE_CLIP_RELATIVE: str = 'timing_clip_relative'
    TIMING_TYPE_CLIP_ABSOLUTE: str = 'timing_clip_absolute'

    def __init__(self, clip, clip_time):
        # type: (MidiClip, float) -> None
        self.clip_time: float = clip_time
        self.clip: MidiClip = clip
        self.timing_type: str = self.TIMING_TYPE_CLIP_RELATIVE

    def get_absolute_start_time(self) -> float:
        if self.timing_type == self.TIMING_TYPE_CLIP_ABSOLUTE:
            time_relative_to_clip = self.clip_time
        elif self.timing_type == self.TIMING_TYPE_CLIP_RELATIVE:
            time_relative_to_clip = self.clip_time * self.clip.length_in_seconds
        else:
            raise ValueError(f'start timing type of {self.timing_type} is not supported (error: 864e96c0)')

        return self.clip.start_time + time_relative_to_clip

    def get_absolute_end_time(self) -> float:
        return self.get_absolute_start_time()



from abc import ABC, abstractmethod
import numpy as np


class WaveProducerInterface(ABC):
    MODE_REPLACE = 'replace'
    MODE_MIX = 'mix'

    @abstractmethod
    def render_to_buffer(self, output_buffer: np.ndarray, sample_rate: int, start_sample: int, mode: str = MODE_REPLACE):
        pass



import unittest

import numpy as np

from py_headless_daw.processing.hybrid.transparent_strategy import Transparent
from py_headless_daw.processing.stream.constant_level import ConstantLevel
from py_headless_daw.processing.stream.stream_gain import StreamGain
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit
from py_headless_daw.schema.wiring import Connector


class RenderingTest(unittest.TestCase):
    def test_fork_and_join(self):
        # +--------+     +--------+
        # |        |     |        |
        # | level1 +--+--> gain1  +--+
        # |        |  |  |        |  | +--------+
        # +--------+  |  +--------+  | |        |
        #             |              +-> mixer  |
        #             |              | |        |
        # +--------+  |  +--------+  | +--------+
        # |        +--+  |        |  |
        # | level2 +-----> gain2  +--+
        # |        |     |        |
        # +--------+     +--------+
        gain1_value = np.float32(0.7)
        gain2_value = np.float32(0.113)

        level1_value = np.float32(0.63)
        level2_value = np.float32(0.47)

        host = Host()

        level1 = Unit(0, 0, 1, 0, host, ConstantLevel(level1_value))
        level2 = Unit(0, 0, 1, 0, host, ConstantLevel(level2_value))

        gain1 = Unit(1, 0, 1, 0, host, StreamGain(gain1_value))
        gain2 = Unit(1, 0, 1, 0, host, StreamGain(gain2_value))

        mixer = Unit(1, 0, 1, 0, host, Transparent())

        Connector(level1.output_stream_nodes[0], gain1.input_stream_nodes[0])
        Connector(level2.output_stream_nodes[0], gain2.input_stream_nodes[0])
        Connector(level2.output_stream_nodes[0], gain1.input_stream_nodes[0])

        Connector(gain1.output_stream_nodes[0], mixer.input_stream_nodes[0])
        Connector(gain2.output_stream_nodes[0], mixer.input_stream_nodes[0])

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1
        interval.start_in_seconds = -0.1
        interval.end_in_seconds = 0.333333333
        interval.num_samples = 44100 * interval.get_length_in_seconds()

        out_buffer = np.ndarray(shape=(100,), dtype=np.float32)
        mixer.output_stream_nodes[0].render(interval, out_buffer)

        expected_output_level = (level1_value + level2_value) * gain1_value + level2_value * gain2_value

        for x in range(0, 100):
            self.assertEqual(expected_output_level, out_buffer[x])



import unittest

import numpy as np

from py_headless_daw.processing.stream.constant_level import ConstantLevel
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit


class ConstantLevelStrategyTest(unittest.TestCase):
    def test_smoke(self):
        host = Host()
        strategy = ConstantLevel(np.float32(0.50))
        unit = Unit(0, 0, 1, 0, host, strategy)

        all_output_nodes = unit.get_all_output_nodes()

        self.assertEqual(1, len(all_output_nodes))

        output_node = all_output_nodes[0]

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1
        interval.start_in_seconds = 0.0
        interval.end_in_seconds = 0.11
        interval.num_samples = 44100 * interval.get_length_in_seconds()

        out_buffer = np.ndarray(shape=(100,), dtype=np.float32)
        output_node.render(interval, out_buffer)

        for x in range(0, 100):
            self.assertEqual(0.5, out_buffer[x])


if __name__ == '__main__':
    unittest.main()



from typing import List

from py_headless_daw.project.having_parameters import HavingParameters


class Track(HavingParameters):
    def __init__(self):
        super().__init__()

        self.outputs = []  # type: List[Track]
        self.inputs = []  # type: List[Track]
        self.name: str = 'unnamed'

    def add_input(self, input_track):
        self.inputs.append(input_track)

    def add_output(self, output_track):
        # type: (Track)->None
        self.outputs.append(output_track)
        output_track.inputs.append(self)



import logging
import os
from typing import List, Optional

from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.project.content.midi_clip import MidiClip
from py_headless_daw.project.content.midi_note import MidiNote
from py_headless_daw.project.envelope import Envelope, EnvelopePoint
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.project.plugins.vst_plugin import VstPlugin
from py_headless_daw.project.project import Project
from py_headless_daw.project.project_renderer import ProjectRenderer
from py_headless_daw.project.sampler_track import SamplerTrack

logger = logging.getLogger(__name__)


class RidingOnBugs:
    """
    A class that generates a demo song called "Riding on Bugs"
    """
    logger = logging.getLogger(__name__)

    def __init__(self, renderer: ProjectRenderer):
        self._renderer: ProjectRenderer = renderer
        self.bar_length: float = 0.3789
        self.length_bars: int = 32

    def render(self, output_file: str):
        logger.info('started rendering RidingOnBugs demo to ' + output_file)

        project = self.create_project()
        self._renderer.render_to_file(project, 0, self.length_bars * self.bar_length, output_file)

    def create_project(self) -> Project:
        bass_drum_track = self._create_bass_drum_track()

        synth_track = self._create_synth_track()
        drum_synth_track = self._create_drum_synth_track()

        master_track = AudioTrack()
        master_track.inputs = [bass_drum_track, synth_track, drum_synth_track]

        return Project(master_track)

    def _create_bass_drum_track(self) -> AudioTrack:
        bass_drum_track = SamplerTrack()
        bass_drum_track.clips = [self._generate_bd_clip(i) for i in range(self.length_bars)]
        reverb = VstPlugin(
            self._get_current_dir() + '/../../../test/test_plugins/DragonflyRoomReverb-vst.x86_64-linux.so')
        bass_drum_track.plugins = [reverb]

        bd_track_gain = bass_drum_track.get_gain_parameter()
        envelope: Envelope = Envelope(bd_track_gain)

        envelope.points = [
            EnvelopePoint(0, 0),
            EnvelopePoint(self.length_bars * self.bar_length / 4, 1.0),
            EnvelopePoint(self.length_bars * self.bar_length / 4 * 3, 1.0),
            EnvelopePoint(self.length_bars * self.bar_length, 0.0)
        ]

        return bass_drum_track

    def _create_drum_synth_track(self) -> AudioTrack:
        drum_midi_track = MidiTrack(1)
        drum_midi_track.clips = self._get_drum_synth_midi_clips()
        drum_midi_track.clips = self._get_drum_synth_midi_clips()
        drum_synth_plugin = DrumSynthPlugin()
        drum_synth_track = AudioTrack()

        drum_synth_track.plugins = [drum_synth_plugin]
        drum_synth_track.inputs = [drum_midi_track]
        drum_synth_track.name = 'drum synth track'

        drum_synth_track.set_parameter_value(drum_synth_track.get_gain_parameter().name, 0.01)

        # todo describe drum synth in doc/drum_synth_plugin.md

        drum_synth_track.set_parameter_value(drum_synth_track.get_panning_parameter().name, -0.9)
        # todo add setting non-default panning just for test
        return drum_synth_track

    def _create_synth_track(self):
        midi_track = MidiTrack(1)
        midi_track.clips = self._get_synth_midi_clips()

        synth_track = AudioTrack()
        synth_plugin = VstPlugin(
            self._get_current_dir() + '/../../../test/test_plugins/amsynth-vst.x86_64-linux.so')

        synth_track.plugins = [
            synth_plugin
        ]

        synth_track.inputs = [midi_track]
        return synth_track

    def _generate_bd_clip(self, bar_number: int) -> AudioClip:
        start_time = bar_number * self.bar_length
        end_time = start_time + self.bar_length / 2
        res = AudioClip(start_time, end_time, self._get_current_dir() + '/resources/bd_mono.wav', 0, 1.0)
        return res

    @staticmethod
    def _get_current_dir() -> str:
        return str(os.path.dirname(os.path.realpath(__file__)))

    def _get_synth_midi_clips(self) -> List[MidiClip]:
        clips = []
        for i in range(self.length_bars):
            clip = self._generate_synth_midi_clip(i)
            if clip is not None:
                clips.append(clip)

        return clips

    def _get_drum_synth_midi_clips(self) -> List[MidiClip]:
        clips = []
        for i in range(self.length_bars):
            clip = self._generate_synth_midi_clip(i)
            if clip is not None:
                clips.append(clip)

        return clips

    def _generate_synth_midi_clip(self, bar_number: int) -> Optional[MidiClip]:
        if bar_number != 3:
            return None

        start = bar_number * self.bar_length
        end = start + self.bar_length
        clip = MidiClip(start, end)

        note = MidiNote(
            clip=clip,
            clip_time=self.bar_length * 0.25,
            note=65,
            velocity=40,
            length=self.bar_length / 2
        )

        clip.midi_notes = [note]
        return clip



from typing import List

from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.content.midi_clip import MidiClip, MidiClipEvent
from py_headless_daw.project.exceptions import RoutingException
from py_headless_daw.project.project import Track


class MidiTrack(Track):

    def __init__(self, channel: int):
        super().__init__()
        self.channel: int = channel
        self.clips: List[MidiClip] = []

    def add_output(self, output: Track):
        if not isinstance(output, AudioTrack):
            raise RoutingException(
                'output of a midi track should be an audio track, instead attempted to use as output: ' + str(
                    type(output)))
        super().add_output(output)

    def get_events(self, start_time: float, end_time: float) -> List[MidiClipEvent]:
        """
        Importantly, this method can return events that fall out to the left of the given time bracket
        because of the hanging notes:
        See ![image](https://user-images.githubusercontent.com/21345604/84991144-c93cec80-b14e-11ea-9053-a6763de8d1e2.png)


        :param start_time:
        :param end_time:
        :return:
        """

        clips = self._find_overlapping_clips(start_time, end_time)
        res: List[MidiClipEvent] = []

        for clip in clips:
            events = self._find_overlapping_events(clip, start_time, end_time)
            for event in events:
                res.append(event)

        return res

    def _find_overlapping_clips(self, start_time: float, end_time: float) -> List[MidiClip]:
        """
        http://i.imgur.com/HpSAG4P.png

        :param start_time:
        :param end_time:
        :return:
        """
        res: List[MidiClip] = []
        for clip in self.clips:
            overlaps: bool = True
            if clip.end_time < start_time:
                overlaps = False
            if clip.start_time >= end_time:
                # http://i.imgur.com/BJI5ol6.png
                overlaps = False

            if overlaps:
                res.append(clip)

        return res

    @staticmethod
    def _find_overlapping_events(clip: MidiClip, start_time: float, end_time: float) -> List[MidiClipEvent]:
        res: List[MidiClipEvent] = []
        tolerance: float = 0.0000001
        for event in clip.events:
            overlaps: bool = True
            if event.get_absolute_start_time() >= end_time - tolerance:
                overlaps = False
            if event.get_absolute_end_time() < start_time + tolerance:
                overlaps = False
            if overlaps:
                res.append(event)

        return res



import logging
import sys
from pathlib import Path

from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.project.content.midi_clip import MidiClip
from py_headless_daw.project.content.midi_note import MidiNote
from py_headless_daw.project.envelope import Envelope, EnvelopePoint
from py_headless_daw.project.exceptions import RoutingException
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.project.plugins.internal_plugin import GainPlugin
from py_headless_daw.project.plugins.vst_plugin import VstPlugin
from py_headless_daw.project.project import Project
from py_headless_daw.project.project_renderer import ProjectRenderer
from py_headless_daw.project.sampler_track import SamplerTrack
from test.container_aware_test_case import ContainerAwareTestCase


class ProjectTest(ContainerAwareTestCase):

    def test_midi_track_output_target(self):
        with self.assertRaises(RoutingException):
            midi_track = MidiTrack(1)
            other_midi_track = MidiTrack(1)
            # noinspection PyTypeChecker
            midi_track.add_output(other_midi_track)

        audio_track = AudioTrack()
        midi_track.add_output(audio_track)

    # noinspection PyMethodMayBeStatic
    def test_create_project(self):
        """
                                 envelope        envelope
                                    |               |
                                    v               v
                           synth->reverb     rev->(gain)           dsynth
        +------------+    +-------------+    +--------+     +------------------+
        |            |    |             |    |        |     |                  |
        | midi_track +----> synth_track +----> master <-----+ drum_synth_track |
        |            |    |             |    |        |     |                  |
        +------------+    +-----+-------+    +----^---+     +--------^---------+
                                +                 |                  |
                               send1              |                  |
                                |                 |           +------+-----------+
                           +----v-------+         |           |                  |
                           |            |         |           |  drum_midi_track |
                           | send_track +---------+           |                  |
                           |            |         |           +------------------+
                           +----^-------+         |
                                |                 |
                               send2              |
                                |                 |
                          +-----+---------+       |
                          |               |       |
                          | sampler_track +-------+
                          |               |
                          +---------------+

        :return:
        """

        # - tracks
        master_track: AudioTrack = AudioTrack()
        master_track.name = 'master'

        synth_track: AudioTrack = AudioTrack()
        drum_synth_track: AudioTrack = AudioTrack()

        synth_midi_track: MidiTrack = MidiTrack(1)
        drum_midi_track: MidiTrack = MidiTrack(1)

        sampler_track: SamplerTrack = SamplerTrack()
        send_track: AudioTrack = AudioTrack()
        send1: AudioTrack = AudioTrack()
        send2: AudioTrack = AudioTrack()

        synth_track.name = "Synth Track"
        drum_synth_track.name = "Drum Synth Track"

        synth_midi_track.name = "Synth Midi Track"
        drum_midi_track.name = "Drum Midi Track"

        sampler_track.name = "Sampler Track"
        send_track.name = "Send Bus"
        send1.name = "Send1"
        send2.name = "Send2"

        # - commutation
        synth_midi_track.add_output(synth_track)
        synth_track.add_output(master_track)
        synth_track.add_output(send1)
        sampler_track.add_output(master_track)
        sampler_track.add_output(send2)
        send1.add_output(send_track)
        send2.add_output(send_track)
        send_track.add_output(master_track)
        drum_synth_track.add_input(drum_midi_track)
        drum_synth_track.add_output(master_track)

        # - adding some midi content
        midi_clip = MidiClip(0.0, 1.0)
        midi_clip.midi_notes = [
            MidiNote(midi_clip, 0.0, 65, 77, 0.1),
            MidiNote(midi_clip, 0.25, 66, 72, 0.1),
            MidiNote(midi_clip, 0.5, 67, 55, 0.1),
            MidiNote(midi_clip, 0.75, 68, 13, 0.1),
        ]

        midi_clip_for_drum_synth = MidiClip(0.0, 1.0)
        midi_clip_for_drum_synth.midi_notes = [
            MidiNote(midi_clip, 0.1, 65, 77, 0.1),
            MidiNote(midi_clip, 0.35, 66, 72, 0.1),
            MidiNote(midi_clip, 0.6, 67, 55, 0.1),
            MidiNote(midi_clip, 0.78, 68, 13, 0.1),
        ]

        synth_midi_track.clips = [midi_clip]
        drum_midi_track.clips = [midi_clip_for_drum_synth]

        # - adding some audio content
        path_to_file = str(Path(__file__).parents[0]) + "/resources/test.wav"

        clip_constructor_args = {
            "start_time": 0.1,
            "end_time": 1.1,
            "source_file": path_to_file,
            "cue_sample": 40,
            "rate": 1.1
        }

        audio_clip = AudioClip(**clip_constructor_args)
        sampler_track.clips = [audio_clip]

        # - putting plugins to tracks
        synth = self._create_vst_plugin('amsynth-vst.x86_64-linux.so', 'synth')
        effect_on_synth_track = self._create_vst_plugin('DragonflyRoomReverb-vst.x86_64-linux.so',
                                                        'reverb on synth track')
        effect_on_sampler_track = self._create_vst_plugin('DragonflyRoomReverb-vst.x86_64-linux.so',
                                                          'reverb on sampler track')
        effect_on_send_track = self._create_vst_plugin('DragonflyRoomReverb-vst.x86_64-linux.so',
                                                       'reverb on send track')
        effect_on_master_track = self._create_vst_plugin('DragonflyRoomReverb-vst.x86_64-linux.so',
                                                         'reverb on master track')

        synth_track.plugins = [
            synth, effect_on_synth_track
        ]

        sampler_track.plugins = [
            effect_on_sampler_track
        ]

        send_track.plugins = [
            effect_on_send_track
        ]

        master_track.plugins = [
            effect_on_master_track
        ]

        drum_synth_track.plugins = [
            DrumSynthPlugin()
        ]

        master_volume_envelope = Envelope(master_track.get_parameter(GainPlugin.PARAMETER_GAIN))

        point11 = EnvelopePoint(0.0, 0.0)
        point12 = EnvelopePoint(0.0, 1.0)

        master_volume_envelope.points = [point11, point12]

        reverb_envelope = Envelope(effect_on_synth_track.get_parameter('Early Level'))

        point21 = EnvelopePoint(0.0, 0.0)
        point22 = EnvelopePoint(0.0, 1.0)

        reverb_envelope.points = [point21, point22]

        project = Project(master_track)

        renderer: ProjectRenderer = self.get_container().project_renderer()

        renderer.render_to_array(project, 0, 10)

    @staticmethod
    def _create_vst_plugin(so_name: str, plugin_name: str = "no name") -> VstPlugin:
        dir_of_this_file = str(Path(__file__).parents[0])
        res = VstPlugin(
            dir_of_this_file + '/../test_plugins/' + so_name)
        res.name = plugin_name
        return res


logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)



from typing import List, Callable

import numpy as np

from py_headless_daw.project.value_provider_consumer import ValueProvider
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class ValueProviderBasedEventEmitter(ProcessingStrategy):
    DEFAULT_SAMPLING_PERIOD = 0.05

    def __init__(self, provider: ValueProvider, transformer: Callable[[float, int], Event],
                 sampling_period=DEFAULT_SAMPLING_PERIOD):
        super().__init__()
        self._provider: ValueProvider = provider
        self._transformer_function: Callable[[float, int], Event] = transformer
        self._sampling_period = sampling_period

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        # warning! dummy implementation, to be refined
        # it's just emitting an event in the beginning of every block
        value: float = self._provider.get_value(interval.start_in_seconds)
        event: Event = self._transformer_function(value, 0)
        event_outputs[0].append(event)



import logging
from typing import List, Union, Dict, cast, Optional, Sequence

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.exceptions import SchemaException
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.wiring import StreamNode, EventNode, Node

RENDERING_TYPE_REPLACE: str = 'replace'
RENDERING_TYPE_ADD: str = 'add'


class Unit:

    def __init__(self, number_of_stream_inputs: int, number_of_event_inputs: int, number_of_stream_outputs: int,
                 number_of_event_outputs: int, host: Host, processing_strategy=None):
        """

        :param number_of_stream_inputs:
        :param number_of_event_inputs:
        :param number_of_stream_outputs:
        :param number_of_event_outputs:
        :param processing_strategy:
        """
        self.host = host

        self._input_stream_nodes: List[StreamNode] = [StreamNode(self) for _ in range(number_of_stream_inputs)]
        self._output_stream_nodes: List[StreamNode] = [StreamNode(self) for _ in range(number_of_stream_outputs)]

        self._input_event_nodes: List[EventNode] = [EventNode(self) for _ in range(number_of_event_inputs)]
        self._output_event_nodes: List[EventNode] = [EventNode(self) for _ in range(number_of_event_outputs)]

        self.last_processed_interval_id: int = -1

        # inplace import to avoid cycle reference
        from py_headless_daw.schema.processing_strategy import ProcessingStrategy
        self._processing_strategy: Optional["ProcessingStrategy"] = None

        if processing_strategy is not None:
            self.set_processing_strategy(processing_strategy)

        self.name = 'unnamed'
        self._internal_buffers: Dict[Node, np.ndarray] = {}

    @property
    def output_stream_nodes(self) -> List[StreamNode]:
        return self._output_stream_nodes

    @property
    def input_stream_nodes(self) -> List[StreamNode]:
        return self._input_stream_nodes

    @property
    def output_nodes(self) -> List[Node]:
        return cast(List[Node], self._output_stream_nodes) + cast(List[Node], self._output_event_nodes)

    @property
    def input_nodes(self) -> List[Node]:
        return cast(List[Node], self._input_stream_nodes) + cast(List[Node], self._input_event_nodes)

    @property
    def output_event_nodes(self) -> List[EventNode]:
        return self._output_event_nodes

    @property
    def input_event_nodes(self) -> List[EventNode]:
        return self._input_event_nodes

    def set_processing_strategy(self, strategy):
        self._processing_strategy = strategy
        strategy.unit = self

    def get_processing_strategy(self):
        return self._processing_strategy

    def _refresh_internal_buffers(self, interval: TimeInterval, size_of_stream_buffer: Union[int, None] = None):
        """

        :param interval:
        :param size_of_stream_buffer:
        :return:
        """

        if interval.id == self.last_processed_interval_id:
            # already processed
            return

        if size_of_stream_buffer is not None:
            self._allocate_internal_stream_buffers(size_of_stream_buffer)

        self._allocate_internal_event_buffers()

        for stream_node in self.input_stream_nodes:
            stream_node.render(interval, self._find_internal_buffer(stream_node))

        for event_node in self.input_event_nodes:
            event_node.render(interval, self._find_internal_buffer(event_node))

        self.last_processed_interval_id = interval.id
        logging.debug('rendering processing strategy for unit "' + self.name + '" on %d:(%2.2f,%2.2f,%d)' % (
            interval.id, interval.start_in_seconds, interval.end_in_seconds, interval.num_samples))

        if self._processing_strategy is None:
            raise Exception('processing strategy not initialized in this unit')

        self._processing_strategy.render(interval, self._get_input_stream_buffers(), self._get_output_stream_buffers(),
                                         self._get_input_event_buffers(), self._get_output_event_buffers())

    def get_num_input_units(self) -> int:

        """

        finds a number of distinct units sending something into this one

        :return:
        """
        found_units: Dict[Unit, int] = {}
        all_input_nodes: List[Node] = []
        all_input_nodes.extend(self.input_stream_nodes)
        all_input_nodes.extend(self.input_event_nodes)

        return len(found_units.keys())

    def render(self, interval: TimeInterval, out_stream_buffer: Union[np.ndarray, None],
               out_event_buffer: Union[List[Event], None],
               out_node: Node, rendering_type: str):
        """

        :param rendering_type:
        :param interval:
        :param out_stream_buffer:
        :param out_event_buffer:
        :param out_node:
        :return:
        """

        self._validate_interval(interval)

        if out_stream_buffer is not None:
            size_of_stream_buffer: int = len(out_stream_buffer)
            self._refresh_internal_buffers(interval, size_of_stream_buffer)
        else:
            self._refresh_internal_buffers(interval)

        if out_node.is_stream():
            self._render_stream_buffer_to_output(
                cast(StreamNode, out_node),
                out_stream_buffer,
                rendering_type)

        if out_event_buffer is not None:
            if out_node.is_event():
                self._render_event_buffer_to_output(
                    cast(EventNode, out_node),
                    out_event_buffer,
                    rendering_type
                )

    def _render_stream_buffer_to_output(self, out_node: StreamNode, out_stream_buffer: np.ndarray, rendering_type: str):
        """

        :param out_node:
        :param out_stream_buffer:
        :return:
        """
        internal_buffer: np.ndarray = self._find_internal_buffer(out_node)
        if rendering_type == RENDERING_TYPE_REPLACE:
            np.copyto(out_stream_buffer, internal_buffer)
        elif rendering_type == RENDERING_TYPE_ADD:
            out_stream_buffer += internal_buffer
        else:
            raise SchemaException('unknown rendering type "' + rendering_type + '"')

    def _render_event_buffer_to_output(self, out_node: EventNode, out_event_buffer: List[Event], rendering_type: str):
        """

        :param out_node:
        :param out_event_buffer:
        :return:
        """
        internal_buffer: List[Event] = self._find_internal_buffer(out_node)
        for event in internal_buffer:
            out_event_buffer.append(event)

    def _find_internal_buffer(self, node: Node) -> Union[np.ndarray, List[Event]]:
        """

        :param node:
        :return:
        """
        return self._internal_buffers[node]

    def _allocate_internal_stream_buffers(self, size_of_stream_buffer: int):
        """

        :param size_of_stream_buffer:
        :return:
        """
        for node in self.input_stream_nodes + self.output_stream_nodes:

            if node in self._internal_buffers:
                # if size of the existing buffer does not match the requested size, we
                # forget this buffer
                existing_shape = self._internal_buffers[node].shape

                if existing_shape != (size_of_stream_buffer,):
                    self._internal_buffers[node] = None

            if node not in self._internal_buffers:
                self._internal_buffers[node] = np.ndarray([size_of_stream_buffer], np.float32)

    def _allocate_internal_event_buffers(self):
        """

        :return:
        """
        for node in self.input_event_nodes + self.output_event_nodes:
            self._internal_buffers[node] = []

    def _get_input_stream_buffers(self) -> List[np.ndarray]:
        return self._convert_nodes_into_buffers(self.input_stream_nodes)

    def _get_output_stream_buffers(self) -> List[np.ndarray]:
        return self._convert_nodes_into_buffers(self.output_stream_nodes)

    def _get_input_event_buffers(self) -> List[List[Event]]:
        return self._convert_nodes_into_buffers(self.input_event_nodes)

    def _get_output_event_buffers(self) -> List[List[Event]]:
        return self._convert_nodes_into_buffers(self.output_event_nodes)

    def _convert_nodes_into_buffers(self, nodes: Sequence[Node]) -> List:
        res = []
        for node in nodes:
            res.append(self._find_internal_buffer(node))
        return res

    def is_output(self, node_to_check: Node) -> bool:
        for node in self.get_all_output_nodes():
            if node_to_check is node:
                return True
        return False

    def is_input(self, node_to_check: Node) -> bool:
        for node in self.get_all_input_nodes():
            if node_to_check == node:
                return True
        return False

    def get_all_output_nodes(self) -> Sequence[Node]:
        # noinspection PyTypeChecker
        return self.output_stream_nodes + self.output_event_nodes  # type: ignore

    def get_all_input_nodes(self) -> Sequence[Node]:
        # noinspection PyTypeChecker
        return self.input_stream_nodes + self.input_event_nodes  # type: ignore

    def _validate_interval(self, interval: TimeInterval):
        inferred_sample_rate = round(interval.num_samples / (interval.end_in_seconds - interval.start_in_seconds))
        host_sample_rate = self.host.sample_rate

        if inferred_sample_rate != host_sample_rate:
            raise Exception(f"""sample rate inferred from the intrval: {inferred_sample_rate}, does not match
the host sample rate {host_sample_rate} (error: 3fd4b60f)""")



from typing import List

from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator_config import DrumSynthGeneratorConfig


class DrumSynthConfigFactory:
    def get_all_preset_names(self) -> List[str]:
        pass

    def randomize(self, config: DrumSynthGeneratorConfig, temperature: float):
        pass



from typing import List

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class ConstantLevel(ProcessingStrategy):

    def __init__(self, level: np.float32):
        super().__init__()
        self.level: np.float32 = level

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        for buffer in stream_outputs:
            buffer.fill(self.level)



from typing import Tuple

from cython_vst_loader.vst_plugin import VstPlugin as RealVstPlugin, VstHost

from py_headless_daw.project.parameter import Parameter
from py_headless_daw.project.plugins.plugin import Plugin


class VstPlugin(Plugin):
    def __init__(self, path_to_shared_library: str, param_value_range: Tuple[float, float] = (0.0, 1.0)):
        super().__init__()

        path_as_bytes: bytes = path_to_shared_library.encode('utf-8')
        # this plugin instance will not be actually used for any processing
        # so it does not really matter which sample frequency and block size we assign to this
        # host
        tmp_host: VstHost = VstHost(44100, 512)
        plugin_instance: RealVstPlugin = RealVstPlugin(path_as_bytes, tmp_host)

        self.path_to_shared_library: str = path_to_shared_library
        self.is_synth: bool = plugin_instance.is_synth()

        for i in range(0, plugin_instance.get_num_parameters()):
            name: str = plugin_instance.get_parameter_name(i).decode()
            value: float = plugin_instance.get_parameter_value(i)
            # on vst parameter ranges: https://www.kvraudio.com/forum/viewtopic.php?t=359767
            self.add_parameter(name, value, Parameter.TYPE_FLOAT, param_value_range)



from random import Random
from typing import Dict, TypeVar

T = TypeVar('T')
RandomChoiceOptions = Dict[T, int]


class Seed:
    """
    A controllable random seed for all kinds of repeatable random uses

    Requirements:
      - similar seeds should produce similar outputs
      - ...
    """

    def __init__(self, seed: str = None):
        self._seed: str = 'default seed'
        if seed is not None:
            self._seed = seed
        self._generator: Random = Random(seed)

    def randint(self, min_val: int, max_val: int, sub_seed: str):
        local_generator = Random(self._seed + sub_seed)
        return local_generator.randint(min_val, max_val)

    def randfloat(self, min_val: float, max_val: float, sub_seed: str) -> float:
        local_generator = Random(self._seed + sub_seed)
        rnd = local_generator.random()
        return min_val + rnd * (max_val - min_val)

    def choose_one(self, probabilities: RandomChoiceOptions[T], sub_seed: str) -> T:
        sum_of_probabilities: int = 0

        # probabilities do not have to sum up to 100 or anything,
        # it's enough if they are positive integers
        for option, add_probability in probabilities.items():
            sum_of_probabilities += add_probability

        dice_value = self.randint(0, sum_of_probabilities - 1, sub_seed)

        sum_of_probabilities = 0
        for option, add_probability in probabilities.items():
            if sum_of_probabilities <= dice_value < sum_of_probabilities + add_probability:
                return option
            sum_of_probabilities += add_probability

        raise ValueError('unable to generate a random choice (error: 8256cd68)')



from abc import ABC
from typing import List, Sequence

from py_headless_daw.project.content.clip import Clip
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.processing_strategy import ProcessingStrategy
from py_headless_daw.shared.clip_intersection import ClipIntersection


class ClipTrackProcessingStrategy(ProcessingStrategy, ABC):
    """
    This is a base class for all strategies that act on the number of clips.
    Examples:
        - midi track (midi data kept in clips)
        - sampler track (audio data kept in clips)

    """

    def __init__(self, clips: Sequence[Clip]):
        self._clips: Sequence[Clip] = clips
        super().__init__()

    def _find_intersections(self, interval: TimeInterval) -> List[ClipIntersection]:
        res: List[ClipIntersection] = []

        for clip in self._clips:
            intersection = ClipIntersection.create_intersection_of_clip_and_interval(clip, interval)
            if intersection is not None:
                res.append(intersection)

        return res



from typing import List

import numpy as np
import random

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class ImpulseNoiseGeneratorStrategy(ProcessingStrategy):
    """
    generates portions of noise intermittenly with portions of silence.
    Used mainly for testing and initial development.
    """

    def __init__(self):
        super().__init__()
        self.times_called = 0

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        num_samples = stream_outputs[0].shape[0]

        for i in range(0, num_samples):
            if i < num_samples // 2 and self.times_called < 2:
                for output in stream_outputs:
                    output[i] = random.uniform(-1, 1)
            else:
                for output in stream_outputs:
                    output[i] = .0

        self.times_called += 1



from abc import ABC, abstractmethod
from typing import List, Optional


class ValueProvider(ABC):
    """
    A provider can send values to multiple consumers
    """

    def __init__(self):
        self.outputs = []  # type: List[ValueConsumer]

    @abstractmethod
    def get_value(self, time: float) -> float:
        pass


class ValueConsumer:
    """
    A consumer can have 0 or 1 providers
    """

    def __init__(self):
        self._value_provider: Optional[ValueProvider] = None

    @property
    def value_provider(self) -> Optional[ValueProvider]:
        return self._value_provider

    @value_provider.setter
    def value_provider(self, provider: ValueProvider):
        self._value_provider = provider
        provider.outputs.append(self)



from typing import Optional

from py_headless_daw.project.content.clip import Clip
from py_headless_daw.schema.dto.time_interval import TimeInterval


class ClipIntersection:
    def __init__(self):
        self.clip: Optional[Clip] = None
        self.start_clip_time: Optional[float] = None
        self.end_clip_time: Optional[float] = None
        self.start_project_time: Optional[float] = None
        self.end_project_time: Optional[float] = None

    @staticmethod
    def create_intersection_of_clip_and_interval(clip, interval):
        """
        This is just a stub to be overridden with an implementation below in this module.
        This is to avoid cyclic dependency (class referring to itself)

        For the same reason, type annotations are done here in this docstring instead of normal hints.

        :param clip
        :type clip Clip

        :param interval
        :type interval TimeInterval

        :rtype: ClipIntersection
        """
        pass


def _create_intersection_of_clip_and_interval(clip: Clip, interval: TimeInterval) -> Optional[ClipIntersection]:
    """
    returns an intersection object or None if the clip and the interval do not intersect
    """

    if clip.end_time < interval.start_in_seconds:
        return None

    if clip.start_time > interval.end_in_seconds:
        return None

    res = ClipIntersection()

    res.start_project_time = max(interval.start_in_seconds, clip.start_time)
    res.end_project_time = min(interval.end_in_seconds, clip.end_time)

    res.start_clip_time = max(float(0), interval.start_in_seconds - clip.start_time)
    res.end_clip_time = min(clip.end_time - clip.start_time, interval.end_in_seconds - clip.start_time)

    res.clip = clip

    return res


ClipIntersection.create_intersection_of_clip_and_interval = _create_intersection_of_clip_and_interval  # type: ignore



from dependency_injector import providers
from dependency_injector.containers import DeclarativeContainer

from py_headless_daw.compiler.project_compiler import InternalPluginProcessingStrategyFactory, ProjectCompiler
from py_headless_daw.project.project_renderer import ProjectRenderer
from py_headless_daw.services.wave_data_provider import WaveDataProvider


class ApplicationContainer(DeclarativeContainer):
    waveform_provider = providers.Singleton(
        WaveDataProvider
    )

    internal_plugin_processing_strategy_factory = providers.Singleton(
        InternalPluginProcessingStrategyFactory,
        waveform_provider
    )

    project_compiler = providers.Singleton(
        ProjectCompiler,
        internal_plugin_processing_strategy_factory=internal_plugin_processing_strategy_factory
    )

    project_renderer = providers.Singleton(
        ProjectRenderer,
        project_compiler=project_compiler
    )



import numpy as np

from py_headless_daw.dsp_utils.curve import get_curve_value
from py_headless_daw.dsp_utils.wave_producer_interface import WaveProducerInterface


# https://en.wikipedia.org/wiki/Envelope_(music)#ADSR
class ADSREnvelope(WaveProducerInterface):

    def __init__(self, attack_time: float, decay_time: float, sustain_level: float, sustain_time: float,
                 release_time: float):
        self.attack_time: float = attack_time
        self.decay_time: float = decay_time
        self.sustain_level: float = sustain_level
        self.sustain_time: float = sustain_time
        self.release_time: float = release_time

        self.attack_curve: float = 0  # 0 means linear, other values in [-1, 1] range (will) be different extents of curveness
        self.decay_curve: float = 0
        self.release_curve: float = 0

        self.attack_curve_power: float = 2.0
        self.decay_curve_power: float = 2.0
        self.release_curve_power: float = 3.0

    def render_to_buffer(self, output_buffer: np.ndarray, sample_rate: int, start_sample: int, mode: str = WaveProducerInterface.MODE_REPLACE):
        for i in range(output_buffer.shape[0]):
            output_buffer[i] = self.get_one_value(start_sample + i, sample_rate)

    def get_one_value(self, sample: int, sample_rate: int) -> float:
        end_of_attack_sample: int = round(self.attack_time * sample_rate)
        end_of_decay_sample: int = end_of_attack_sample + round(self.decay_time * sample_rate)
        end_of_sustain_sample: int = end_of_decay_sample + round(self.sustain_time * sample_rate)
        end_of_release_sample: int = end_of_sustain_sample + round(self.release_time * sample_rate)

        if 0 <= sample <= end_of_attack_sample:
            if end_of_attack_sample == 0:
                return 1.0

            phase = sample / end_of_attack_sample
            return get_curve_value(0, 1.0, self.attack_curve, self.attack_curve_power, phase)
        elif end_of_attack_sample < sample <= end_of_decay_sample:
            if sample == end_of_decay_sample:
                return self.sustain_level
            phase = (sample - end_of_attack_sample) / (end_of_decay_sample - end_of_attack_sample)
            return get_curve_value(1.0, self.sustain_level, self.decay_curve, self.decay_curve_power, phase)
        elif end_of_decay_sample < sample <= end_of_sustain_sample:
            if sample == end_of_sustain_sample:
                return self.sustain_level
            return self.sustain_level
        elif end_of_sustain_sample < sample <= end_of_release_sample:
            if sample == end_of_release_sample:
                return 0.0

            phase = (sample - end_of_sustain_sample) / (end_of_release_sample - end_of_sustain_sample)
            return get_curve_value(self.sustain_level, 0, self.release_curve, self.release_curve_power, phase)
        else:
            return 0.0

    @staticmethod
    def _get_curve_value(start_val: float, end_val: float, curve_ratio: float, curve_power: float, phase: float):
        linear_value = start_val + (end_val - start_val) * phase
        if curve_ratio >= 0:
            curve_value = start_val + (end_val - start_val) * (phase ** curve_power)
        else:
            curve_value = start_val + (end_val - start_val) * (1 - (1 - phase) ** curve_power)

        abs_ratio = abs(curve_ratio)

        return linear_value * (1 - abs_ratio) + curve_value * abs_ratio

    def get_length_seconds(self) -> float:
        return self.attack_time + self.decay_time + self.sustain_time + self.release_time



from typing import Union

from py_headless_daw.schema.events.event import Event


class ParameterValueEvent(Event):

    @property
    def type(self) -> str:
        return Event.TYPE_PARAMETER_VALUE

    def __init__(self, sample_position: int, parameter_id: str, value: Union[float, str]):
        super().__init__(sample_position)
        self.parameter_id: str = parameter_id
        self.value: Union[float, str] = value



import unittest
from pathlib import Path

import parametrized

from py_headless_daw.application_container import ApplicationContainer
from py_headless_daw.production.seed import Seed
from py_headless_daw.production.simple_edm.simple_edm_producer import SimpleEdmProducer
from py_headless_daw.project.project_renderer import ProjectRenderer


@parametrized.parametrized_test_case
class SimpleEdmProducerTest(unittest.TestCase):

    @parametrized.parametrized([
        ['hello'], ['hello1'], ['hello2'], ['123']
    ])
    def generate_audio(self, seed_string: str):
        seed = Seed(seed_string)
        producer = SimpleEdmProducer(seed)
        project = producer.generate_project()

        container = ApplicationContainer()

        compiler = container.project_compiler()
        renderer = ProjectRenderer(compiler)

        out_file_path = str(Path(__file__).parent.absolute()) + f"/output/out_{seed_string}.wav"

        if project.start_time_seconds is None:
            raise ValueError()

        if project.end_time_seconds is None:
            raise ValueError()

        renderer.render_to_file(project, project.start_time_seconds, project.end_time_seconds, out_file_path)

        pass


if __name__ == '__main__':
    unittest.main()



import os
import unittest

from py_headless_daw.production.simple_edm.amsynth_patches_manager import AmsynthPatchesManager
from py_headless_daw.project.named_parameter_bag import NamedParameterBag


class AmsynthPatchesManagerTest(unittest.TestCase):
    def test_getting_all_patches(self):
        manager = AmsynthPatchesManager(self._get_patches_dir())

        all_patches = manager.get_all_patches()
        self.assertEqual(787, len(all_patches))

    def test_getting_patches_of_one_group(self):
        manager = AmsynthPatchesManager(self._get_patches_dir())
        patches = manager.get_all_patches_from_group('PatriksBank02.bank.txt')
        self.assertEqual(127, len(patches))
        self.assertIsInstance(patches[0], NamedParameterBag)

    def _get_patches_dir(self) -> str:
        this_dir_path = os.path.dirname(os.path.realpath(__file__))
        return this_dir_path + '/../../../py_headless_daw/production/simple_edm/amsynth_patches/'



class Clip:
    def __init__(self, start_time: float, end_time: float):
        self.start_time: float = start_time
        self.end_time: float = end_time

    @property
    def length_in_seconds(self) -> float:
        return self.end_time - self.start_time



import unittest

import numpy as np

from py_headless_daw.processing.hybrid.drum_synth_strategy import DrumSynthStrategy
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.midi_event_factory import MidiEventFactory
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit
# import matplotlib.pyplot as plt


class DrumSynthStrategyTestCase(unittest.TestCase):
    def test_render_one_buffer_with_a_note(self):
        plugin = DrumSynthPlugin()

        host = Host()
        strategy = DrumSynthStrategy(plugin)
        Unit(0, 0, 1, 0, host, strategy)

        out_stream_buffer = np.zeros(shape=(100,), dtype=np.float32)

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1

        factory = MidiEventFactory()

        note_event = factory.create_note_on_event(note=65, velocity=70, position=50)

        strategy.render(interval, [], [out_stream_buffer], [[note_event]], [])

        zero_sample_indices = [0, 20, 40]
        non_zero_sample_indices = [60, 80, 99]

        for idx in zero_sample_indices:
            self.assertEqual(0, out_stream_buffer[idx])

        for idx in non_zero_sample_indices:
            self.assertNotEqual(0, out_stream_buffer[idx])


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import List

import numpy as np

from py_headless_daw.processing.stream.stream_gain import StreamGain
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.parameter_value_event import ParameterValueEvent


class StreamGainStrategyTest(unittest.TestCase):

    def test_stream_gain_strategy(self):
        strategy = StreamGain(np.float32(0.25))

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1

        in_stream_buffer = np.ones(shape=(100,), dtype=np.float32)
        out_stream_buffer = np.zeros(shape=(100,), dtype=np.float32)

        setter_event: ParameterValueEvent = ParameterValueEvent(0, StreamGain.PARAMETER_GAIN, 0.55)

        input_event_buffer: List[Event] = [setter_event]
        output_event_buffer: List[Event] = []

        strategy.render(interval, [in_stream_buffer], [out_stream_buffer], [input_event_buffer], [output_event_buffer])

        # the first few samples are closer to the initial value
        for x in range(0, 3):
            self.assertTrue(0.24 < out_stream_buffer[x] < 0.26)

        # while the last few are closer to the target one
        for x in range(out_stream_buffer.shape[0] - 3, out_stream_buffer.shape[0]):
            self.assertTrue(0.24 < out_stream_buffer[x] > 0.45)

        strategy.render(interval, [in_stream_buffer], [out_stream_buffer], [[]], [[]])
        # now we render without any events in the input, the logic in the
        # strategy is slightly different in this case
        for x in range(out_stream_buffer.shape[0] - 3, out_stream_buffer.shape[0]):
            self.assertTrue(out_stream_buffer[x] > 0.45)



import unittest

import numpy as np

from py_headless_daw.dsp_utils.adsr_envelope import ADSREnvelope
from py_headless_daw.dsp_utils.drum_synth.one_shot_oscillator import OneShotOscillator


# noinspection PyUnresolvedReferences
# import matplotlib.pyplot as plt


class OneShotOscillatorTestCase(unittest.TestCase):
    def test_get_value(self):
        envelope = ADSREnvelope(
            attack_time=0.1,
            decay_time=0.4,
            sustain_level=0.5,
            sustain_time=0.25,
            release_time=0.25
        )

        osc = OneShotOscillator()

        osc.pitch_envelope = envelope
        osc.volume_envelope = envelope

        output_buffer_size = 1000
        sample_rate = 44100

        output_buffer = np.ndarray(shape=(output_buffer_size,), dtype=np.float32)
        osc.render_to_buffer(output_buffer, sample_rate, 0)

        # plt.plot(output_buffer)
        # plt.show()


if __name__ == '__main__':
    unittest.main()



import numpy as np  # type: ignore


class Waveform:
    def __init__(self, sample_rate: int, data: np.ndarray):
        self.sample_rate: int = sample_rate
        self.data: np.ndarray = data
        pass

    def length_in_samples(self) -> int:
        return self.data.shape[1]

    @property
    def num_channels(self) -> int:
        return self.data.shape[0]



import numpy as np
import math

from py_headless_daw.dsp_utils.adsr_envelope import ADSREnvelope
from py_headless_daw.dsp_utils.wave_producer_interface import WaveProducerInterface


class OneShotOscillator(WaveProducerInterface):
    # different signals https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sawtooth.html
    TYPE_SINE: str = "sine"
    TYPE_NOISE: str = "noise"
    TYPE_TRIANGLE: str = "triangle"

    def __init__(self):
        self.pitch_envelope: ADSREnvelope = self._create_default_envelope()
        self.volume_envelope: ADSREnvelope = self._create_default_envelope()
        self.zero_frequency: float = 440.0
        self.frequency_range: float = 440.0
        self.wave = self.TYPE_SINE
        self.distortion: float = 0.0
        self.initial_phase: float = 0.0
        self.volume: float = 1.0

    def render_to_buffer(self, output_buffer: np.ndarray, sample_rate: int, start_sample: int,
                         mode: str = WaveProducerInterface.MODE_REPLACE):

        if output_buffer.ndim != 1:
            raise ValueError(f"buffers given to an oscillator are supposed to be mono, "
                             f"this one has {output_buffer.ndim} dimensions instead of 1")

        volume_env_wave = np.copy(output_buffer)
        self.volume_envelope.render_to_buffer(volume_env_wave, sample_rate, start_sample)

        oscillation_wave = np.copy(output_buffer)
        self._generate_oscillation_wave(oscillation_wave, sample_rate, start_sample)

        np.multiply(oscillation_wave, volume_env_wave, out=oscillation_wave)

        if WaveProducerInterface.MODE_REPLACE == mode:
            np.copyto(output_buffer, oscillation_wave)
        elif WaveProducerInterface.MODE_MIX == mode:
            np.add(output_buffer, oscillation_wave, out=output_buffer)
        else:
            raise ValueError(f'unknown producing mode {mode}')

    def _create_default_envelope(self) -> ADSREnvelope:
        return ADSREnvelope(attack_time=0, decay_time=1, sustain_level=1, sustain_time=0, release_time=1)

    def _generate_oscillation_wave(self, out_buffer: np.ndarray, sample_rate, start_sample):
        phase: float = self.initial_phase
        for i in range(out_buffer.shape[0]):
            relative_frequency = self.pitch_envelope.get_one_value(start_sample + i, sample_rate)
            real_frequency = self.zero_frequency + self.frequency_range * relative_frequency

            out_buffer[i] = self._wave_function(self.wave, self.distortion, phase)

            if real_frequency > 0:
                samples_in_period = (1 / real_frequency) * sample_rate
                oscillator_step_for_one_sample = 2 * math.pi / samples_in_period
                phase += oscillator_step_for_one_sample

    def _wave_function(self, waveform: str, distortion: float, phase: float) -> float:
        return math.sin(phase)



from typing import List

from py_headless_daw.schema.unit import Unit
from py_headless_daw.schema.wiring import StreamNode, EventNode


class Chain(Unit):
    """
    represents a linear chain of units which is a unit on its own

    """

    # we dont need to call parent constructor in this case
    # noinspection PyMissingConstructor
    def __init__(self, first_unit: Unit, last_unit: Unit):
        """
        Units in the chain are expected to be properly interconnected prior to be passed here

        :param first_unit:
        :param last_unit:
        """

        if isinstance(first_unit, Unit) is not True:
            raise Exception

        if isinstance(last_unit, Unit) is not True:
            raise Exception

        self._fist_unit: Unit = first_unit
        self._last_unit: Unit = last_unit

    @Unit.input_stream_nodes.getter  # type: ignore
    def input_stream_nodes(self) -> List[StreamNode]:
        return self._fist_unit.input_stream_nodes

    @Unit.input_event_nodes.getter  # type: ignore
    def input_event_nodes(self) -> List[EventNode]:
        return self._fist_unit.input_event_nodes

    @Unit.output_stream_nodes.getter  # type: ignore
    def output_stream_nodes(self) -> List[StreamNode]:
        return self._last_unit.output_stream_nodes

    @Unit.output_event_nodes.getter  # type: ignore
    def output_event_nodes(self) -> List[EventNode]:
        return self._last_unit.output_event_nodes



from typing import Optional

from py_headless_daw.project.track import Track


class Project:
    def __init__(self, master_track: Track):
        self.start_time_seconds: Optional[float] = None
        self.end_time_seconds: Optional[float] = None

        self.master_track: Track = master_track
        self.num_audio_channels = 2
        pass



import unittest

from py_headless_daw.project.audio_track import AudioTrack


class TrackTest(unittest.TestCase):
    def test_audio_track(self):
        track = AudioTrack()
        self.assertEqual(2, len(track.parameters))
        self.assertEqual(2, len(track.plugins))



import unittest
import subprocess
import os


class LintingTest(unittest.TestCase):
    def test_flake8_main_code_and_tests(self):
        this_dir = os.path.dirname(os.path.realpath(__file__))

        checks_to_ignore = [
            'E501',
            'W503'
        ]

        cmd_line_parts = [
            "flake8",
            this_dir + "/../py_headless_daw",
            this_dir + "/../test",
            "--count",
            f'--ignore={",".join(checks_to_ignore)}',
            '--show-source',
            '--statistics'
        ]

        result = subprocess.run(cmd_line_parts)
        self.assertEqual(0, result.returncode, str(result.stdout) + str(result.stderr))

    def test_mypy_main_code_and_tests(self):
        this_dir = os.path.dirname(os.path.realpath(__file__))

        for source_relative_dir in ["/../py_headless_daw/", ""]:
            cmd_line_parts = [
                "mypy",
                "--ignore-missing-imports",
                this_dir + source_relative_dir,
            ]

            result = subprocess.run(cmd_line_parts)
            self.assertEqual(0, result.returncode, str(result.stdout) + str(result.stderr))

    def test_mypy_test_code(self):
        this_dir = os.path.dirname(os.path.realpath(__file__))

        cmd_line_parts = [
            "mypy",
            "--ignore-missing-imports",
            this_dir + "/../py_headless_daw/",
        ]

        result = subprocess.run(cmd_line_parts)
        self.assertEqual(0, result.returncode, str(result.stdout) + str(result.stderr))


if __name__ == '__main__':
    unittest.main()



from abc import abstractmethod, ABC
from typing import List, Optional

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.unit import Unit


class ProcessingStrategy(ABC):

    def __init__(self):
        self._unit: Optional[Unit] = None

    @abstractmethod
    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        pass

    @property
    def unit(self) -> Optional[Unit]:
        return self._unit

    @unit.setter
    def unit(self, unit: Unit):
        self._unit = unit

    @staticmethod
    def _flatten_event_inputs(event_inputs: List[List[Event]]) -> List[Event]:

        events: List[Event] = []

        # we gather all midi events from all event inputs
        for event_input in event_inputs:
            for event in event_input:
                events.append(event)

        return events

    def _filter_events(self, event_inputs: List[List[Event]], event_type: str) -> List[Event]:
        flattened_events: List[Event] = self._flatten_event_inputs(event_inputs)

        res = []
        for event in flattened_events:
            if event.type == event_type:
                res.append(event)

        return res



from typing import cast, List

from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator_config import DrumSynthGeneratorConfig, OscillatorConfig
from py_headless_daw.project.parameter import Parameter
from py_headless_daw.project.plugins.internal_plugin import InternalPlugin
import inspect

DEFAULT_FREQUENCY_RANGE = 20_000


class DrumSynthPlugin(InternalPlugin):
    MAX_NUMBER_OF_OSCILLATORS = 4
    DEFAULT_ZERO_FREQUENCY = 10

    PARAM_VALUE_WAVEFORM_SINE = 'sine'
    PARAM_VALUE_WAVEFORM_SAWTOOTH = 'sawtooth'
    PARAM_VALUE_WAVEFORM_SQUARE = 'square'
    PARAM_VALUE_WAVEFORM_TRIANGLE = 'triangle'
    PARAM_VALUE_WAVEFORM_NOISE = 'noise'

    PARAM_NAME_NUM_OSCILLATORS = 'num_oscillators'

    PARAM_NAME_SUFFIX_OSCILLATOR_WAVEFORM = 'waveform'
    PARAM_NAME_SUFFIX_OSCILLATOR_WAVE_DISTORTION = 'wave_distortion'
    PARAM_NAME_SUFFIX_OSCILLATOR_ZERO_FREQUENCY = 'zero_frequency'
    PARAM_NAME_SUFFIX_OSCILLATOR_FREQUENCY_RANGE = 'frequency_range'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME = 'volume'

    PARAM_NAME_SUFFIX_OSCILLATOR_INITIAL_PHASE = 'initial_phase'

    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_TIME = 'volume_attack_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_CURVE_RATIO = 'volume_attack_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_CURVE_POWER = 'volume_attack_curve_power'

    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_TIME = 'volume_decay_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_CURVE_RATIO = 'volume_decay_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_CURVE_POWER = 'volume_decay_curve_power'

    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_LEVEL = 'sustain_level'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_TIME = 'sustain_time'

    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_TIME = 'volume_release_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_CURVE_RATIO = 'volume_release_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_CURVE_POWER = 'volume_release_curve_power'

    # pitch envelope parameters
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_TIME = 'pitch_attack_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_CURVE_RATIO = 'pitch_attack_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_CURVE_POWER = 'pitch_attack_curve_power'

    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_TIME = 'pitch_decay_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_CURVE_RATIO = 'pitch_decay_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_CURVE_POWER = 'pitch_decay_curve_power'

    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_LEVEL = 'pitch_sustain_level'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_TIME = 'pitch_sustain_time'

    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_TIME = 'pitch_release_time'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_CURVE_RATIO = 'pitch_release_curve_ratio'
    PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_CURVE_POWER = 'pitch_release_curve_power'

    def __init__(self):

        super().__init__()

        self.add_parameter(name=self.PARAM_NAME_NUM_OSCILLATORS,
                           value="1",
                           param_type=Parameter.TYPE_ENUM,
                           value_range=[
                               "1", "2", "3", "4"
                           ])

        for i in range(1, self.MAX_NUMBER_OF_OSCILLATORS + 1):

            self.add_parameter(
                name=self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_WAVEFORM, i),
                value=self.PARAM_VALUE_WAVEFORM_SINE,
                param_type=Parameter.TYPE_ENUM,
                value_range=[
                    self.PARAM_VALUE_WAVEFORM_SINE,
                    self.PARAM_VALUE_WAVEFORM_SAWTOOTH,
                    self.PARAM_VALUE_WAVEFORM_SQUARE,
                    self.PARAM_VALUE_WAVEFORM_TRIANGLE,
                    self.PARAM_VALUE_WAVEFORM_NOISE
                ]
            )

            self.add_parameter(
                name=self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_ZERO_FREQUENCY, i),
                value=self.DEFAULT_ZERO_FREQUENCY,
                param_type=Parameter.TYPE_FLOAT,
                value_range=(0, 22050)
            )

            self.add_parameter(
                name=self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_FREQUENCY_RANGE, i),
                value=DEFAULT_FREQUENCY_RANGE,
                param_type=Parameter.TYPE_FLOAT,
                value_range=(0, 22050)
            )

            for member_name, prop in inspect.getmembers(type(self)):
                member_name = cast(str, member_name)
                if member_name.startswith('PARAM_NAME_SUFFIX_OSCILLATOR'):
                    param_name_suffix = getattr(self, member_name)
                    param_name = self.generate_param_name(param_name_suffix, i)
                    if not self.has_parameter(param_name):
                        value_range = (0.0, 1.0)
                        value = 0.5
                        self.add_parameter(param_name, value, Parameter.TYPE_FLOAT, value_range)

    def generate_generator_config(self) -> DrumSynthGeneratorConfig:
        osc_configs: List[OscillatorConfig] = []
        for i in range(1, int(self.get_parameter_value(self.PARAM_NAME_NUM_OSCILLATORS)) + 1):
            osc_config = OscillatorConfig(
                volume=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME, i)),
                zero_frequency=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_ZERO_FREQUENCY, i)),
                frequency_range=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_FREQUENCY_RANGE, i)),
                start_phase=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_INITIAL_PHASE, i)),
                wave_type=self.get_enum_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_WAVEFORM, i)),

                volume_envelope_attack_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_TIME, i)),
                volume_envelope_attack_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_CURVE_RATIO, i)),
                volume_envelope_attack_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_CURVE_POWER, i)),

                volume_envelope_decay_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_TIME, i)),
                volume_envelope_decay_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_CURVE_RATIO, i)),
                volume_envelope_decay_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_CURVE_POWER, i)),

                volume_envelope_sustain_level=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_LEVEL, i)),
                volume_envelope_sustain_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_TIME, i)),

                volume_envelope_release_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_TIME, i)),
                volume_envelope_release_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_CURVE_RATIO,
                                             i)),
                volume_envelope_release_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_CURVE_POWER,
                                             i)),

                pitch_envelope_attack_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_TIME, i)),
                pitch_envelope_attack_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_CURVE_RATIO, i)),
                pitch_envelope_attack_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_CURVE_POWER, i)),

                pitch_envelope_decay_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_TIME, i)),
                pitch_envelope_decay_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_CURVE_RATIO, i)),
                pitch_envelope_decay_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_CURVE_POWER, i)),

                pitch_envelope_sustain_level=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_LEVEL, i)),
                pitch_envelope_sustain_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_TIME, i)),

                pitch_envelope_release_time=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_TIME, i)),
                pitch_envelope_release_curve_ratio=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_CURVE_RATIO,
                                             i)),
                pitch_envelope_release_curve_power=self.get_float_parameter_value(
                    self.generate_param_name(self.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_CURVE_POWER,
                                             i))
            )

            osc_configs.append(osc_config)

        res = DrumSynthGeneratorConfig(osc_configs)

        return res

    @staticmethod
    def generate_param_name(suffix: str, oscillator_id: int) -> str:
        return suffix + '_osc' + str(oscillator_id)



from py_headless_daw.schema.events.event import Event


class MidiEvent(Event):

    MIDI_TYPE_REGULAR = 1
    MIDI_TYPE_SYSEX = 2

    @property
    def type(self):
        return self.TYPE_MIDI

    def __init__(self, sample_position: int):
        Event.__init__(self, sample_position)

        self.midi_event_type: int = self.MIDI_TYPE_REGULAR
        self.status: int = 0
        self.data1: int = 0
        self.data2: int = 0
        self.note: int = -1
        self.channel: int = 1

    def is_note_off(self) -> bool:
        # we consider any event with velocity == 0 as note off, which can be not correct sometimes,
        # but is enough for now
        if self.data2 == 0:
            return True

        return False

    def is_note_on(self) -> bool:
        # in fact it's much more complex than this, but it's ok for now
        return self.data2 != 0

    @property
    def velocity(self) -> int:
        return self.data2

    @velocity.setter
    def velocity(self, new_velocity: int):
        self.data2 = new_velocity

    @property
    def note(self) -> int:
        return self.data1

    @note.setter
    def note(self, new_note: int):
        self.data1 = new_note



import unittest

import parametrized

from py_headless_daw.dsp_utils.curve import get_curve_value


@parametrized.parametrized_test_case
class MyTestCase(unittest.TestCase):

    @parametrized.parametrized([
        [0.0, 1.0, 0, 2.0, 0, 0],
        [0.0, 1.0, 0, 2.0, 1, 1],

        [0.0, 1.0, 0.5, 2.0, 0.0, 0],
        [0.0, 1.0, 0.5, 2.0, 1.0, 1],
        [0.0, 1.0, 0.5, 2.0, 0.5, 0.375],  # 0.5 / 2 + 0.5 * 0.5 / 2

        [0.0, 1.0, -0.5, 2.0, 0.0, 0],
        [0.0, 1.0, -0.5, 2.0, 1.0, 1],
        [0.0, 1.0, -0.5, 2.0, 0.5, 0.625]
    ])
    def one_value(self, start_val: float, end_val: float, curve_ratio: float,
                  curve_power: float, phase: float,
                  expected_value: float):
        self.assertEquals(expected_value, get_curve_value(
            start_val=start_val,
            end_val=end_val,
            curve_ratio=curve_ratio,
            curve_power=curve_power,
            phase=phase))


if __name__ == '__main__':
    unittest.main()



from em.platform.rendering.processing_strategies.stream.impulse_noice_generator import ImpulseNoiseGeneratorStrategy
from em.platform.rendering.schema.unit import Unit


class ImpulseNoiseGenerator(Unit):
    def __init__(self):
        processing_strategy = ImpulseNoiseGeneratorStrategy()

        super(ImpulseNoiseGenerator, self).__init__(0, 0, 2, 0, processing_strategy)



import unittest

import numpy as np

from py_headless_daw.processing.stream.constant_level import ConstantLevel
from py_headless_daw.processing.stream.stream_gain import StreamGain
from py_headless_daw.schema.chain import Chain
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit
from py_headless_daw.schema.wiring import Connector


class ChainTest(unittest.TestCase):
    def test_chaining_two_stream_units(self):
        host = Host()
        unit1 = Unit(0, 0, 1, 0, host, ConstantLevel(np.float32(0.5)))
        unit2 = Unit(1, 0, 1, 0, host, StreamGain(np.float32(0.5)))
        Connector(unit1.output_stream_nodes[0], unit2.input_stream_nodes[0])

        chain = Chain(unit1, unit2)

        self.assertEqual(len(chain.input_stream_nodes), 0)
        self.assertEqual(len(chain.input_event_nodes), 0)
        self.assertEqual(len(chain.output_stream_nodes), 1)
        self.assertEqual(len(chain.output_event_nodes), 0)

        interval = TimeInterval()
        interval.start_in_bars = 0
        interval.end_in_bars = 1

        interval.start_in_seconds = 0.0
        interval.end_in_seconds = 0.25

        interval.num_samples = 44100 * interval.get_length_in_seconds()
        out_buffer = np.ndarray(shape=(100,), dtype=np.float32)
        chain.output_stream_nodes[0].render(interval, out_buffer)

        for x in range(0, 100):
            self.assertEqual(0.25, out_buffer[x])



import unittest
from typing import cast

from py_headless_daw.processing.event.midi_track_strategy import MidiTrackStrategy
from py_headless_daw.project.content.midi_clip import MidiClip
from py_headless_daw.project.content.midi_note import MidiNote
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.midi_event import MidiEvent


class MidiTrackStrategyTest(unittest.TestCase):
    def test_something(self):
        track = self._generate_midi_track()
        strategy = MidiTrackStrategy(track)

        outputs = [[]]

        interval = TimeInterval()
        interval.start_in_seconds = 0.0
        interval.end_in_seconds = 1.0
        interval.start_in_bars = 0.0
        interval.end_in_bars = 1.0
        interval.num_samples = 100

        strategy.render(interval, [], [], [], outputs)

        output = outputs[0]

        self.assertEqual(2, len(output))  # note on and note off

        first_event: MidiEvent = cast(MidiEvent, output[0])
        second_event: MidiEvent = cast(MidiEvent, output[1])

        self.assertIsInstance(first_event, MidiEvent)
        self.assertIsInstance(second_event, MidiEvent)

        self.assertEqual(0, first_event.sample_position)
        self.assertEqual(10, second_event.sample_position)

    @staticmethod
    def _generate_midi_track() -> MidiTrack:
        track = MidiTrack(1)
        clip = MidiClip(0.0, 1.0)
        clip.midi_notes = [
            MidiNote(clip, 0.0, 65, 65, 0.1)
        ]
        track.clips = [
            clip
        ]
        return track



from typing import List, Tuple, Union, cast

from py_headless_daw.project.value_provider_consumer import ValueConsumer

ParameterRangeType = Union[Tuple[float, float], List[str]]
ParameterValueType = Union[float, str]


class Parameter(ValueConsumer):
    TYPE_FLOAT = 'float'
    TYPE_ENUM = 'enum'

    def __init__(self, name: str, value: ParameterValueType, parameter_type: str,
                 parameter_range: ParameterRangeType):
        super().__init__()
        self.name: str = name
        if parameter_type not in self.get_available_types():
            raise ValueError(
                f"unsupported parameter type {parameter_type}, supported: {', '.join(self.get_available_types())}")

        self.type: str = parameter_type
        self.range = parameter_range
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value: ParameterValueType):
        if self.type == self.TYPE_FLOAT:
            value_range = cast('Tuple[float, float]', self.range)
            new_value = cast(float, new_value)

            if not isinstance(new_value, float) and not isinstance(new_value, int):
                raise ValueError(
                    f"a new value for parameter {self.name} should have been float, "
                    f"instead is {str(type(new_value))} with value {str(new_value)} (error: 90a2a0ed)")

            if not value_range[0] <= new_value <= value_range[1]:
                raise ValueError(f"float parameter {self.name} value of {new_value} "
                                 f"is out of range {value_range[0]} - {value_range[1]} "
                                 f"(error: 8e343a6e)")

        elif self.type == self.TYPE_ENUM:
            casted_value_range = cast(List[str], self.range)
            if type(new_value) is not str:
                raise ValueError(
                    f"a new value for parameter {self.name} should have been str, "
                    f"instead is {str(type(new_value))} (error: 90a2a0ed)")

            if new_value not in casted_value_range:

                raise ValueError(
                    f"{new_value}, a value for parameter {self.name} should have been one of {', '.join(casted_value_range)}")

        self._value: ParameterValueType = new_value

    @classmethod
    def get_available_types(cls) -> List[str]:
        return [cls.TYPE_FLOAT, cls.TYPE_ENUM]

    @property
    def range(self) -> ParameterRangeType:
        return self._range

    @range.setter
    def range(self, value: ParameterValueType):
        if self.type == self.TYPE_FLOAT:
            if isinstance(value, tuple):
                if len(value) != 2:
                    raise ValueError('ranges for float params must be tuples of length 2 (error: a27c90e2)')
                for i in range(2):
                    if not isinstance(value[i], float) and not isinstance(value[i], int):
                        raise ValueError('ranges for float params must be tuples of floats or ints (error: 80d5db89)')
            else:
                raise ValueError('for float params, range must be a tuple of two floats (error: e37e6f14)')
        elif self.type == self.TYPE_ENUM:
            if isinstance(value, list):
                for e in value:
                    if not isinstance(e, str):
                        raise ValueError('for enum params, each element range must be a string (error: f0dc465c)')
            else:
                raise ValueError('for enum params, range must be a list of strings (error: a9c603f3)')
        else:
            raise ValueError('unsupported type (error: 2878e15c)')

        self._range: ParameterRangeType = value



def get_curve_value(start_val: float, end_val: float, curve_ratio: float, curve_power: float, phase: float):
    """
    this is a simple function for producing transition from start_val to end_val with
    different amounts of skewness

    curve_ratio of 0.0 means linear transition.

    anything above 0 up to 1 means slow start and speeding up towards the end,
    1.0 being the extreme case when no linear component is present at all

    the range [-1.0..0] is the exact opposite to the above.

    the curve_power is the power of the x**power function to get the calculate the skewed
    component
    """
    linear_value = start_val + (end_val - start_val) * phase
    if curve_ratio >= 0:
        curve_value = start_val + (end_val - start_val) * (phase ** curve_power)
    else:
        curve_value = start_val + (end_val - start_val) * (1 - (1 - phase) ** curve_power)

    abs_ratio = abs(curve_ratio)

    return linear_value * (1 - abs_ratio) + curve_value * abs_ratio



from typing import List, Union

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.exceptions import SchemaException


class Connector:
    _registry = []  # type: List[Connector]

    def __init__(self, in_node, out_node):
        # type: (Node, Node) -> None
        self.input_node: Node = in_node
        self.out_node: Node = out_node

        assert self.connected(in_node, out_node) is False, "these two nodes already have a connector between them"

        self.input_node.attach_to_connector_input(self)
        self.out_node.attach_to_connector_output(self)

        self._registry.append(self)

    @classmethod
    def connected(cls, node1, node2):
        # type: (Node, Node)->bool
        for connector in cls._registry:
            if connector.input_node is node1 and connector.out_node is node2:
                return True
            if connector.out_node is node1 and connector.input_node is node2:
                return True

        return False


class Node:
    def __init__(self, processing_unit=None):
        """
        :param processing_unit: # Unit we are not doing type checking here to avoid circular dependency
        """
        from py_headless_daw.schema.unit import Unit
        self.unit: Unit = processing_unit
        self._connectors: List[Connector] = []

    def attach_to_connector_output(self, connector: Connector):
        if self._is_unit_output():
            raise SchemaException(
                'this node is an output of a unit, but it has been attempted to connect to an output of a connector')

        self._connectors.append(connector)

    def attach_to_connector_input(self, connector: Connector):
        if self._is_unit_input():
            raise SchemaException(
                'this node is an input of a unit, but it has been attempted to connect to an input of a connector')

        self._connectors.append(connector)

    def is_stream(self) -> bool:
        return False

    def is_event(self) -> bool:
        return False

    def render(self, interval: TimeInterval, out_buffer: Union[np.ndarray, List[Event]]):
        """

        :param interval:
        :param out_buffer:
        :return:
        """
        # to postpone importing to avoid a cycle:
        from py_headless_daw.schema.unit import RENDERING_TYPE_REPLACE, RENDERING_TYPE_ADD

        nodes_to_render: List[Node]

        if not self._is_unit_output():
            # this node is not directly an output of a unit, so we
            # need to find its outputting units through the connectors
            nodes_to_render = []
            for connector in self._connectors:
                nodes_to_render.append(connector.input_node)
        else:
            # the node IS an output of a unit, so we render it directly`
            nodes_to_render = [self]

        for sequence_number, node_to_render in enumerate(nodes_to_render):

            if self.is_stream():
                if 0 == sequence_number:
                    node_to_render.unit.render(interval, out_buffer, None, node_to_render, RENDERING_TYPE_REPLACE)
                else:
                    node_to_render.unit.render(interval, out_buffer, None, node_to_render, RENDERING_TYPE_ADD)
            if self.is_event():
                if 0 == sequence_number:
                    node_to_render.unit.render(interval, None, out_buffer, node_to_render, RENDERING_TYPE_REPLACE)
                else:
                    node_to_render.unit.render(interval, None, out_buffer, node_to_render, RENDERING_TYPE_ADD)

    def _is_unit_output(self) -> bool:
        if self.unit is not None:
            return self.unit.is_output(self)
        else:
            return False

    def _is_unit_input(self) -> bool:
        return not self._is_unit_output()


class StreamNode(Node):

    def is_stream(self):
        return True


class EventNode(Node):

    def is_event(self):
        return True



from typing import List

from py_headless_daw.project.having_parameters import HavingParameters
from py_headless_daw.project.parameter import Parameter, ParameterValueType, ParameterRangeType
from py_headless_daw.project.plugins.internal_plugin import GainPlugin, PanningPlugin
from py_headless_daw.project.plugins.plugin import Plugin
from py_headless_daw.project.project import Track


class AudioTrack(Track):
    GAIN: str = 'gain'
    PANNING: str = 'panning'

    def __init__(self):
        super().__init__()
        self._plugins: List[Plugin] = []
        self._gain: GainPlugin = GainPlugin()
        self._panner: PanningPlugin = PanningPlugin()

    @HavingParameters.parameters.getter  # type: ignore
    def parameters(self):
        return self._gain.parameters + self._panner.parameters

    def add_parameter(self, name: str,
                      value: ParameterValueType,
                      param_type: str,
                      value_range: ParameterRangeType):
        raise Exception('adding parameters to this object is not supported')

    def get_gain_parameter(self) -> Parameter:
        return self._gain.get_parameter(GainPlugin.PARAMETER_GAIN)

    def get_panning_parameter(self) -> Parameter:
        return self._panner.get_parameter(PanningPlugin.PARAMETER_PANNING)

    @property
    def plugins(self) -> List[Plugin]:
        return self._plugins + [
            self._gain,
            self._panner
        ]

    @plugins.setter
    def plugins(self, plugins: List[Plugin]):
        self._plugins = []
        for plugin in plugins:
            self._plugins.append(plugin)



from typing import List

from py_headless_daw.project.value_provider_consumer import ValueProvider, ValueConsumer


class EnvelopePoint:
    SLOPE_TYPE_LINEAR = 'linear'
    SLOPE_TYPE_QUADRATIC = 'quadratic'
    SLOPE_TYPE_SPLINE = 'spline'

    def __init__(self, time: float, value: float):
        self.time: float = time
        self.value: float = value

        self.left_slope_type: str = self.SLOPE_TYPE_LINEAR
        self.left_slope_parameter: float = 1.1

        self.right_slope_type: str = self.SLOPE_TYPE_LINEAR
        self.right_slope_parameter: float = 1.1


class Envelope(ValueProvider):

    def __init__(self, value_consumer: ValueConsumer):
        super().__init__()
        self.target: ValueConsumer = value_consumer
        self.target.value_provider = self
        self._points: List[EnvelopePoint] = []

    @property
    def points(self) -> List[EnvelopePoint]:
        return self._points

    @points.setter
    def points(self, points: List[EnvelopePoint]):
        self._points = sorted(points, key=lambda point: point.time)

    def get_value(self, time: float) -> float:
        if time <= self._points[0].time:
            return self._points[0].value
        elif time >= self._points[-1].time:
            return self._points[-1].value

        for idx in range(0, len(self._points) - 1):

            prev_time: float = self._points[idx].time
            next_time: float = self._points[idx + 1].time

            prev_value: float = self._points[idx].value
            next_value: float = self._points[idx + 1].value

            if prev_time <= time < next_time:
                return prev_value + (next_value - prev_value) * (time - prev_time) / (next_time - prev_time)

        raise Exception('unable to find a range')



import unittest

from py_headless_daw.production.seed import Seed


class SeedTestCase(unittest.TestCase):
    def test_choose_one(self):
        seed = Seed('hello world')
        choice = seed.choose_one({
            'hello': 10
        }, '1')
        self.assertEqual('hello', choice)

    def test_randint(self):
        seed = Seed('seed1')

        res = seed.randint(0, 10000, "subseed1")
        self.assertEqual(4872, res)

        res = seed.randint(0, 10000, "subseed1")
        self.assertEqual(4872, res)

        res = seed.randint(0, 10000, "subseed2")
        self.assertEqual(6248, res)

        seed = Seed('seed2')

        res = seed.randint(0, 10000, "subseed1")
        self.assertEqual(898, res)

    def test_randfloat(self):
        seed = Seed('123')
        rnd = seed.randfloat(10, 20, 'ss')
        self.assertEqual(12.784257244353906, rnd)
        rnd = seed.randfloat(10, 20, 'ss')
        self.assertEqual(12.784257244353906, rnd)

        rnd = seed.randfloat(10, 20, 'ss1')
        self.assertEqual(16.743403597407195, rnd)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, Union

from py_headless_daw.production.seed import Seed
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.project.plugins.plugin_preset import PluginPreset


class SimpleEdmDrumSynthPresetFactory:

    def generate_bd_preset(self, seed: Seed) -> PluginPreset:
        res = PluginPreset('standard bd', DrumSynthPlugin)

        global_param_data = {
            DrumSynthPlugin.PARAM_NAME_NUM_OSCILLATORS: "1"
        }

        oscillator_param_data: Dict[str, Union[float, str]] = {
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME: 1.0,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_WAVEFORM: DrumSynthPlugin.PARAM_VALUE_WAVEFORM_SINE,

            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_ATTACK_TIME: 0.001,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_DECAY_TIME: seed.randfloat(0.05, 0.06,
                                                                                                    "bd decay time"),
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_LEVEL: 0.0,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_SUSTAIN_TIME: 0.0,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_VOLUME_ENVELOPE_RELEASE_TIME: 0.0,

            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_ATTACK_TIME: 0.01,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_DECAY_TIME: seed.randfloat(0.05, 0.06,
                                                                                                   "bd pitch decay time"),
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_LEVEL: 0.0,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_SUSTAIN_TIME: 0.0,
            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_PITCH_ENVELOPE_RELEASE_TIME: 0.0,

            DrumSynthPlugin.PARAM_NAME_SUFFIX_OSCILLATOR_FREQUENCY_RANGE: 200.0,
        }

        for name, value in global_param_data.items():
            res.set_parameter_value(name, value)

        for name_prefix in oscillator_param_data:
            # some weird things below are made to make mypy happy
            osc_value: Union[str, float] = oscillator_param_data[name_prefix]
            param_name = DrumSynthPlugin.generate_param_name(name_prefix, 1)
            res.set_parameter_value(param_name, osc_value)

        return res



from pathlib import Path

from py_headless_daw.application_container import ApplicationContainer
from py_headless_daw.production.seed import Seed
from py_headless_daw.production.simple_edm.simple_edm_producer import SimpleEdmProducer
from py_headless_daw.project.project_renderer import ProjectRenderer

seed = Seed('hello1')
producer = SimpleEdmProducer(seed)
project = producer.generate_project()

container = ApplicationContainer()

compiler = container.project_compiler()
renderer = ProjectRenderer(compiler)

out_file_path = str(Path(__file__).parent.absolute()) + "/output/out.wav"

if project.start_time_seconds is None:
    raise ValueError()

if project.end_time_seconds is None:
    raise ValueError()

renderer.render_to_file(project, project.start_time_seconds, project.end_time_seconds, out_file_path)



from typing import List, cast

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.parameter_value_event import ParameterValueEvent
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class StreamGain(ProcessingStrategy):
    PARAMETER_GAIN: str = 'gain'

    def __init__(self, gain: np.float32):
        self.gain: np.float32 = gain
        super().__init__()

    # noinspection PyShadowingNames
    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        # simplified approach:
        # we check all events arrived and react on any param value events,
        # the last one we face will determine current gain value, if none have arrived,
        # the gain will not change
        all_events = self._flatten_event_inputs(event_inputs)
        all_events = sorted(all_events, key=lambda event: event.sample_position)

        new_desired_gain = self.gain

        for event in all_events:
            if isinstance(event, ParameterValueEvent):
                event = cast(ParameterValueEvent, event)
                if event.parameter_id == self.PARAMETER_GAIN:
                    new_desired_gain = event.value

        if new_desired_gain != self.gain:
            # even in this simplified approach,
            # we still want to avoid abrupt volume changes resulting in clicks,
            # for that we smoothen out transition:
            # if a change in gain is detected (new desired gain != current gain),
            # it will be reached during this buffer by applying a slowly changing array
            gain_operation = self._get_smoothener_array(self.gain, new_desired_gain, stream_outputs[0].shape[0])
        else:
            gain_operation = self.gain

        self.gain = new_desired_gain

        for i in range(0, len(stream_inputs)):
            np.copyto(stream_outputs[i], stream_inputs[i])
            stream_outputs[i] *= gain_operation

    def _get_smoothener_array(self, initial_gain: float, new_desired_gain: float, num_samples: int) -> np.ndarray:
        return np.linspace(start=initial_gain, stop=new_desired_gain, num=num_samples, dtype=np.float32)



class Host:

    def __init__(self):
        self._bpm: float = 120.0
        self._block_size: int = 1024
        self._sample_rate: int = 44100

    @property
    def block_size(self) -> int:
        return self._block_size

    @block_size.setter
    def block_size(self, new_size: int):
        self._block_size = new_size

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, new_sample_rate: int):
        self._sample_rate = new_sample_rate

    @property
    def bpm(self) -> float:
        return self._bpm

    @bpm.setter
    def bpm(self, new_bpm: float):
        self._bpm = new_bpm



from typing import cast

from py_headless_daw.shared.having_name import HavingName
from py_headless_daw.project.having_parameters import HavingParameters


class Plugin(HavingParameters, HavingName):
    def configure_with_preset(self, preset) -> None:
        from py_headless_daw.project.plugins.plugin_preset import PluginPreset  # to avoid a cyclic load time dependency
        preset = cast(PluginPreset, preset)
        for parameter in preset.parameters:
            val_from_preset = preset.get_parameter_value(parameter.name)
            self.set_parameter_value(parameter.name, val_from_preset)



class TimeInterval:
    id_cursor: int = 0

    def __init__(self):
        self.id = self.id_cursor
        TimeInterval.id_cursor += 1
        self.start_in_bars: float = 0
        self.end_in_bars: float = 0
        self.num_samples: int = 0
        self._start_in_seconds: float = 0
        self._end_in_seconds: float = 0

    def get_length_in_bars(self) -> float:
        return self.end_in_bars - self.start_in_bars

    def get_length_in_seconds(self) -> float:
        return self.end_in_seconds - self.start_in_seconds

    @property
    def start_in_seconds(self) -> float:
        if self._start_in_seconds is None:
            raise Exception('start in seconds is not set for this interval')
        return self._start_in_seconds

    @start_in_seconds.setter
    def start_in_seconds(self, new: float):
        self._start_in_seconds = new

    @property
    def end_in_seconds(self) -> float:
        if self._end_in_seconds is None:
            raise Exception('start in seconds is not set for this interval')
        return self._end_in_seconds

    @end_in_seconds.setter
    def end_in_seconds(self, new: float):
        self._end_in_seconds = new

    def sample_rate(self) -> int:
        return round(self.num_samples / (self._end_in_seconds - self._start_in_seconds))



from py_headless_daw.schema.events.midi_event import MidiEvent


class MidiEventFactory:
    SAMPLE_POSITION_UNDEFINED: int = -1

    STATUS_TEMPLATE_NOTE_ON: int = 0b10010000
    STATUS_TEMPLATE_NOTE_OFF: int = 0b10000000
    STATUS_TEMPLATE_PROGRAM_CHANGE: int = 0b11000000

    def __init__(self):
        self.channel_number: int = 0
        pass

    def create_note_on_event(self, note: int, velocity: int, position: int = SAMPLE_POSITION_UNDEFINED) -> MidiEvent:
        status_byte: int = self.STATUS_TEMPLATE_NOTE_ON
        status_byte += self.channel_number

        data1: int = note
        data2: int = velocity

        event = MidiEvent(position)

        event.status = status_byte
        event.data1 = data1
        event.data2 = data2

        event.note = note
        return event

    def create_note_off_event(self, note: int, velocity: int, position: int = SAMPLE_POSITION_UNDEFINED) -> MidiEvent:
        return self.create_note_on_event(note=note, velocity=0, position=position)



from py_headless_daw.project.content.clip import Clip


class AudioClip(Clip):
    def __init__(self, start_time: float, end_time: float, source_file: str, cue_sample: int, rate: float):
        super().__init__(start_time, end_time)
        self.source_file_path: str = source_file
        self.cue_sample: int = cue_sample
        self.rate: float = rate
        self.gain: float = 1.0



import unittest

import numpy as np

from py_headless_daw.dsp_utils.adsr_envelope import ADSREnvelope


class ADSREnvelopeTestCase(unittest.TestCase):
    def test_happy_case_get_one_value(self):
        envelope = ADSREnvelope(
            attack_time=1,
            decay_time=1,
            sustain_level=0.5,
            sustain_time=1,
            release_time=1
        )

        self.assertEqual(0, envelope.get_one_value(-1, 1000))
        self.assertEqual(0, envelope.get_one_value(0, 1000))
        self.assertEqual(0.5, envelope.get_one_value(500, 1000))
        self.assertEqual(1, envelope.get_one_value(1000, 1000))
        self.assertEqual(0.75, envelope.get_one_value(1500, 1000))
        self.assertEqual(0.5, envelope.get_one_value(2000, 1000))
        self.assertEqual(0.5, envelope.get_one_value(2500, 1000))
        self.assertEqual(0.5, envelope.get_one_value(3000, 1000))
        self.assertEqual(0.25, envelope.get_one_value(3500, 1000))
        self.assertEqual(0.0, envelope.get_one_value(4000, 1000))
        self.assertEqual(0.0, envelope.get_one_value(4001, 1000))

    def test_generate_wave(self):
        envelope = ADSREnvelope(
            attack_time=1,
            decay_time=1,
            sustain_level=0.5,
            sustain_time=1,
            release_time=1
        )

        buffer = np.ndarray(shape=(400,), dtype=np.float32)
        envelope.render_to_buffer(buffer, 100, 0)

    def test_get_length(self):
        envelope = ADSREnvelope(
            attack_time=1,
            decay_time=1,
            sustain_level=0.5,
            sustain_time=1,
            release_time=1
        )
        self.assertEqual(4.0, envelope.get_length_seconds())


if __name__ == '__main__':
    unittest.main()



from typing import Type

from py_headless_daw.project.having_parameters import HavingParameters
from py_headless_daw.project.plugins.plugin import Plugin


class PluginPreset(HavingParameters):

    def __init__(self, name: str,
                 plugin_class: Type[Plugin]):  # https://docs.python.org/3/library/typing.html#typing.Type
        self.name: str = name
        super().__init__()

        # instantiating a blank plugin instance
        plugin = plugin_class()

        for parameter_from_plugin in plugin.parameters:
            self.add_parameter_object(parameter_from_plugin)



from abc import ABC, abstractmethod
from typing import Tuple

import scipy.io.wavfile
import numpy as np

from py_headless_daw.dto.waveform import Waveform


class WaveformProviderInterface(ABC):

    @abstractmethod
    def get_wave_data_by_file_path(self, file_path: str) -> Waveform:
        pass


class WaveDataProvider(WaveformProviderInterface):
    """
    Gets waveforms from different sources, such as local files.
    The result is a Waveform object.
    """

    def get_wave_data_by_file_path(self, file_path: str) -> Waveform:
        wav_file_details: Tuple[int, np.ndarray] = scipy.io.wavfile.read(file_path)
        (sample_rate, data) = wav_file_details

        if data.ndim > 1:
            data = np.rot90(data)
        else:
            data = data.reshape((1, data.shape[0]))

        divider = None
        if data.dtype is np.dtype(np.int16):
            divider = 32768
        elif data.dtype is np.dtype(np.int32):
            divider = 2147483648

        if divider is not None:
            float_data = data / divider
            data = float_data.astype(np.float32)

        return Waveform(sample_rate, data)



from typing import List

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class StereoPanner(ProcessingStrategy):

    def __init__(self, panning: float):
        super().__init__()
        self.panning = panning

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        left_gain: float = self._get_left_gain(self.panning)
        right_gain: float = self._get_right_gain(self.panning)

        for i in range(0, len(stream_inputs)):
            np.copyto(stream_outputs[i], stream_inputs[i])

        stream_outputs[0] *= left_gain
        stream_outputs[1] *= right_gain

    def _get_left_gain(self, gain: float):
        if gain < 0:
            return 1

        return 1 - gain

    def _get_right_gain(self, gain: float):
        if gain > 0:
            return 1

        return gain + 1.0



from typing import List, Dict

import numpy as np

from em.platform.rendering.dto.time_interval import TimeInterval
from em.platform.rendering.schema.events.event import Event
from em.platform.rendering.schema.processing_strategy import ProcessingStrategy


class Mixer(ProcessingStrategy):

    def __init__(self):
        self.gains: Dict[int, float] = {}
        self.pannings: Dict[int, float] = {}

    def set_gain(self, track_number: int, gain: float):
        self.gains[track_number] = gain

    def get_gain(self, track_number: int):
        return self.gains[track_number]

    def set_panning(self, track_number: int, panning: float):
        self.pannings[track_number] = panning

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):

        num_input_stream_channels: int = len(stream_inputs)
        num_output_stream_channels: int = len(stream_outputs)

        for stream_output in stream_outputs:
            stream_output.fill(0.0)

        # assuming each unit provides the same number of channels to the mixer,
        # if it doesn't, everything will break
        num_input_units: int = int(num_input_stream_channels / num_output_stream_channels)
        for u in range(0, num_input_units):

            base_gain: float = self._get_gain(u)
            panning: float = self._get_panning(u)

            left_gain: float = base_gain * (1 - max([0, panning]))
            right_gain: float = base_gain * (1 + min([0, panning]))

            stream_outputs[0] += left_gain * stream_inputs[u * num_output_stream_channels + 0]
            stream_outputs[1] += right_gain * stream_inputs[u * num_output_stream_channels + 1]

    def _get_gain(self, u) -> float:
        if u in self.gains:
            return self.gains[u]
        else:
            return 1.0

    def _get_panning(self, u) -> float:
        if u in self.pannings:
            return self.pannings[u]
        else:
            return 0.0



from typing import List

from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.project.parameter import Parameter
from py_headless_daw.project.plugins.plugin import Plugin


class InternalPlugin(Plugin):
    pass


class GainPlugin(InternalPlugin):
    PARAMETER_GAIN: str = 'gain'

    def __init__(self):
        super().__init__()
        self.add_parameter(self.PARAMETER_GAIN, 1.0, Parameter.TYPE_FLOAT, (0, 1.0))


class PanningPlugin(InternalPlugin):
    PARAMETER_PANNING: str = 'panning'

    def __init__(self):
        super().__init__()
        self.add_parameter(self.PARAMETER_PANNING, 0.0, Parameter.TYPE_FLOAT, (-1.0, 1.0))


class SamplerPlugin(InternalPlugin):

    def __init__(self, clips: List[AudioClip]):
        self.clips: List[AudioClip] = clips
        super().__init__()



import unittest

from py_headless_daw.integrations.amsynth.amsynth_parameter_normalizer import AmsynthParameterNormalizer


class MyTestCase(unittest.TestCase):
    def test_denormalize(self):
        normalizer = AmsynthParameterNormalizer()
        denormalized_val = normalizer.denormalize('amp_attack', 0.059)
        self.assertEqual(0.1475, denormalized_val)

    def test_normalize(self):
        normalizer = AmsynthParameterNormalizer()
        normalized_val = normalizer.normalize('amp_attack', 0.1475)
        self.assertEqual(0.059, normalized_val)


if __name__ == '__main__':
    unittest.main()



import unittest
from typing import List, Tuple, Optional

import numpy as np
import parametrized

from py_headless_daw.dto.waveform import Waveform
from py_headless_daw.processing.stream.sampler import Sampler
from py_headless_daw.project.content.audio_clip import AudioClip
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit
from py_headless_daw.services.wave_data_provider import WaveformProviderInterface


class MockedWaveformProvider(WaveformProviderInterface):

    def get_wave_data_by_file_path(self, file_path: str) -> Waveform:
        """
        interprets the file path as:

        numchannels_volume_lengthinsamples_samplerate, e.g. 2_0.123_100_44100 and generates a flat "wave"
        of the given length (in samples) on a given sample rate

        """

        (num_channels, volume, length_in_samples, sample_rate) = self._parse_file_path(file_path)
        data = np.zeros((num_channels, length_in_samples)) + volume
        return Waveform(sample_rate, data)

    def _parse_file_path(self, file_path: str) -> Tuple[int, float, int, int]:
        elements = file_path.split('_')
        if len(elements) != 4:
            raise Exception('the file path should have had exactly 3 underscore-delimited parts')

        (num_channels, volume, length_in_samples, sample_rate) = (
            int(elements[0]),
            float(elements[1]),
            int(elements[2]),
            int(elements[3])
        )

        return num_channels, volume, length_in_samples, sample_rate


@parametrized.parametrized_test_case
class SamplerTest(unittest.TestCase):

    @parametrized.parametrized([
        [0.0, 1.0, 100, 1.0, None],
        [1.0, 2.0, 100, 0.0, None],
        [2.0, 3.0, 100, 2.0, None],
        [3.0, 4.0, 100, 0.0, None],
        [0.5, 1.5, 100, None, 0.5],
        [0.75, 1.75, 100, None, 0.25],
        [10.0, 11.0, 100, None, 0.5],  # clip goes beyond underlying file
        [13.0, 14.0, 100, 1.0, None],  # single-channel clip
    ])
    def one_case(self, interval_start: float, interval_end: float, buffer_length: int,
                 expected_level_in_output: Optional[float], expected_average: Optional[float]):
        host = Host()
        host.sample_rate = 100

        unit = Unit(2, 0, 2, 0, host)

        clips: List[AudioClip] = self.generate_test_clips()
        mocked_waveform_provider = MockedWaveformProvider()
        strategy = Sampler(clips, mocked_waveform_provider)
        strategy.unit = unit

        interval = TimeInterval()
        interval.start_in_seconds = interval_start
        interval.end_in_seconds = interval_end

        out1 = np.zeros(shape=(buffer_length,), dtype=np.float32)
        out2 = np.zeros(shape=(buffer_length,), dtype=np.float32)

        strategy.render(interval, [], [out1, out2], [], [])

        values_tested: bool = False

        if expected_level_in_output is not None:
            expected_buffer = np.zeros(100, ) + expected_level_in_output
            self.assertTrue((out1 == expected_buffer).all())
            values_tested = True

        if expected_average is not None:
            self.assertEqual(expected_average, np.mean(out1))
            self.assertEquals(expected_average, np.mean(out2))
            values_tested = True

        self.assertTrue(values_tested)

    def generate_test_clips(self) -> List[AudioClip]:
        clip1 = AudioClip(0.0, 1.0, "2_1.0_1000_100", 0, 1.0)

        clip2 = AudioClip(2.0, 3.0, "2_1.0_1000_100", 0, 1.0)
        clip3 = AudioClip(2.0, 3.0, "2_1.0_1000_100", 0, 1.0)

        clip4 = AudioClip(10.0, 11.0, "2_1.0_1000_100", 950, 1.0)

        clip5 = AudioClip(13.0, 14.0, "1_1.0_1000_100", 0, 1.0)

        return [clip1, clip2, clip3, clip4, clip5]



import os


def get_path_relative_to_project_root(relative_path: str) -> str:
    this_dir_path = os.path.dirname(os.path.realpath(__file__))
    return this_dir_path + '/../../' + relative_path


def get_path_relative_to_file(file_name: str, relative_path: str) -> str:
    this_dir_path = os.path.dirname(os.path.realpath(file_name))
    return this_dir_path + '/' + relative_path



import unittest

from py_headless_daw.application_container import ApplicationContainer


class ContainerAwareTestCase(unittest.TestCase):

    def __init__(self, method_name='runTest'):
        self._container = ApplicationContainer()
        super().__init__(method_name)

    def get_container(self) -> ApplicationContainer:
        return self._container



import os
from typing import List, Optional

from py_headless_daw.integrations.amsynth.amsynth_parameter_normalizer import AmsynthParameterNormalizer
from py_headless_daw.production.producer_interface import ProducerInterface
from py_headless_daw.production.seed import Seed
from py_headless_daw.production.simple_edm.amsynth_patches_manager import AmsynthPatchesManager
from py_headless_daw.production.simple_edm.drum_synth_preset_factory import SimpleEdmDrumSynthPresetFactory
from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.content.midi_clip import MidiClip
from py_headless_daw.project.content.midi_note import MidiNote
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.project.named_parameter_bag import NamedParameterBag
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.project.plugins.vst_plugin import VstPlugin
from py_headless_daw.project.project import Project
from py_headless_daw.shared.utils import get_path_relative_to_file


class SimpleEdmProducer(ProducerInterface):

    def __init__(self, seed: Seed):
        self._seed: Seed = seed
        self._bpm: int = self._seed.randint(118, 124, 'bpm')
        self._song_length_beats: int = 32
        self._initial_offset_second = 0.001

    def generate_project(self) -> Project:
        # number_of_synth_tracks = self._invent_number_of_synth_tracks()
        number_of_synth_tracks = 3
        master_track = AudioTrack()

        bd_track_audio_track = self._generate_bd_track()
        master_track.add_input(bd_track_audio_track)

        reverb_send_track = self._generate_reverb_send_track()
        master_track.add_input(reverb_send_track)

        for i in range(0, number_of_synth_tracks):
            synth_track = self._generate_synth_track(i)
            master_track.add_input(synth_track)
            reverb_send_track.add_input(synth_track)

        res = Project(master_track)

        res.start_time_seconds = 0.0
        res.end_time_seconds = 15.0

        return res

    def _invent_number_of_synth_tracks(self) -> int:
        return self._seed.choose_one(
            {
                2: 20,
                3: 20,
                4: 20
            },
            "num synth tracks"
        )

    def _generate_bd_track(self) -> AudioTrack:
        """
        bd track is a midi track feeding itself to drum synth track
        """

        drum_synth_audio_track = self._generate_drum_synth_audio_track_itself()
        drum_synth_midi_track = self._generate_drum_synth_midi_track()

        drum_synth_audio_track.add_input(drum_synth_midi_track)
        return drum_synth_audio_track

    def _generate_reverb_send_track(self) -> AudioTrack:
        res = AudioTrack()

        plugin = self._create_vst_plugin('DragonflyRoomReverb-vst.x86_64-linux.so')
        res.plugins = [plugin]

        plugin.set_parameter_value('Dry Level',
                                   0.00000001)  # for some reason pure 0 seems to randomly give some weird behaviour
        plugin.set_parameter_value('Late Level', 0.5)
        plugin.set_parameter_value('Decay', 0.18)

        return res

    def _create_vst_plugin(self, so_name: str) -> VstPlugin:
        dir_path = os.path.dirname(os.path.realpath(__file__))
        path_to_so: str = dir_path + '/../../../test/test_plugins/' + so_name
        return VstPlugin(path_to_so, (0.0, 1.0))

    def _generate_synth_track(self, i: int) -> AudioTrack:
        synth_midi_track = self._generate_synth_midi_track(i)
        synth_track_itself = self._generate_synth_track_itself(i)
        synth_track_itself.add_input(synth_midi_track)
        return synth_track_itself

    def _generate_drum_synth_audio_track_itself(self) -> AudioTrack:
        res = AudioTrack()
        plugin = DrumSynthPlugin()
        bd_preset = SimpleEdmDrumSynthPresetFactory().generate_bd_preset(self._seed)
        plugin.configure_with_preset(bd_preset)
        res.plugins = [plugin]
        return res

    def _generate_drum_synth_midi_track(self) -> MidiTrack:
        # each clip is 4 beats in length
        res = MidiTrack(channel=1)
        res.clips = [self._generate_one_drum_synth_midi_clip(i) for i in range(0, 4)]
        return res

    def _generate_one_drum_synth_midi_clip(self, clip_number: int) -> MidiClip:
        length_of_one_clip_seconds = 60 / self._bpm * 4

        start_time = clip_number * length_of_one_clip_seconds
        end_time = start_time + length_of_one_clip_seconds

        res = MidiClip(start_time, end_time)

        note1 = MidiNote(res, 0, 65, 80, 0.1)
        note2 = MidiNote(res, 0.25, 65, 80, 0.1)
        note3 = MidiNote(res, 0.5, 65, 80, 0.1)
        note4 = MidiNote(res, 0.75, 65, 80, 0.1)

        res.midi_notes = [note1, note2, note3, note4]
        return res

    def _generate_synth_midi_track(self, i: int) -> MidiTrack:
        res = MidiTrack(1)

        clip = self._generate_synth_clip(i)

        res.clips = [clip]
        return res

    def _generate_synth_track_itself(self, i) -> AudioTrack:
        res = AudioTrack()

        res.set_parameter_value(res.get_gain_parameter().name,
                                self._seed.randfloat(0, 0.5, "gain for synthesized " + str(i)))

        res.set_parameter_value(res.get_panning_parameter().name,
                                self._seed.randfloat(-1.0, 1.0, "panning for synthesized " + str(i)))

        dir_path = os.path.dirname(os.path.realpath(__file__))
        path_to_so: str = dir_path + '/../../../test/test_plugins/amsynth-vst.x86_64-linux.so'

        vst_synth_plugin = VstPlugin(path_to_so, (0.0, 1.0))

        param_bag = self._get_synth_param_bag(i)
        self._apply_params_bag_to_synth_plugin(param_bag, vst_synth_plugin)

        res.plugins = [vst_synth_plugin]
        return res

    def _get_synth_param_bag(self, i) -> NamedParameterBag:
        manager = AmsynthPatchesManager(get_path_relative_to_file(__file__, 'amsynth_patches'))
        all_patches = manager.get_all_patches()
        selected_idx = self._seed.randint(0, len(all_patches), 'synth patch for synth number ' + str(i))
        return all_patches[selected_idx]

    def _apply_params_bag_to_synth_plugin(self, param_bag: NamedParameterBag, vst_synth_plugin: VstPlugin):
        normalizer = AmsynthParameterNormalizer()
        for parameter in param_bag.parameters:
            name = parameter.name
            new_value = normalizer.normalize(name, param_bag.get_float_parameter_value(name))
            vst_synth_plugin.set_parameter_value(name, new_value)

    def _generate_synth_clip(self, synth_id: int) -> MidiClip:
        length_in_bars: int = self._seed.choose_one(
            {
                2: 10,
                4: 10,
                8: 10
            },
            f'pattern length for synth {synth_id}'
        )

        length_in_bars = 8

        length_in_beats = length_in_bars * 4
        length_of_one_beat_seconds = 60 / self._bpm

        behavior_pause = "pause"
        behavior_sustain = "sustain"
        behavior_start_note = "start_note"

        offset = 0.15

        clip: MidiClip = MidiClip(0, length_in_beats * length_of_one_beat_seconds)

        clip.start_time += offset
        clip.end_time += offset

        note_playing: Optional[MidiNote] = None
        notes: List[MidiNote] = []

        for position_id in range(length_in_beats):
            behavior = self._seed.choose_one(
                {
                    behavior_pause: 5,
                    behavior_sustain: 5,
                    behavior_start_note: 10
                },
                f'note behavior for position {position_id} of synth {synth_id}'
            )

            # if position_id % 4 == 0:
            #     behavior = behavior_start_note
            # if position_id % 4 == 1:
            #     behavior = behavior_sustain
            # if position_id % 4 == 2:
            #     behavior = behavior_pause
            # if position_id % 4 == 3:
            #     behavior = behavior_pause

            if note_playing is not None:
                if behavior == behavior_pause:
                    note_playing.length = position_id * length_of_one_beat_seconds - note_playing.clip_time
                    notes.append(note_playing)
                    note_playing = None
                elif behavior in [behavior_start_note, behavior_sustain]:
                    pass
                else:
                    raise ValueError(f'unknown behaviour {behavior} (error: 602bfc62)')
            else:
                if behavior == behavior_start_note:
                    pitch: int = self._seed.randint(50, 60, f"pitch for synth {synth_id} at {position_id}")
                    velocity: int = self._seed.randint(30, 110, f"velocity for synth {synth_id} at {position_id}")
                    # pitch = 55
                    # velocity = 100
                    note_playing = MidiNote(
                        clip,
                        position_id * length_of_one_beat_seconds,
                        pitch,
                        velocity,
                        length_of_one_beat_seconds
                    )
                    note_playing.timing_type = MidiNote.TIMING_TYPE_CLIP_ABSOLUTE
                elif behavior in [behavior_pause, behavior_sustain]:
                    pass
                else:
                    raise ValueError(f'unknown behavior {behavior} (error: 898ee65b')

        clip.midi_notes = notes
        return clip



from typing import List

import numpy as np

from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.processing_strategy import ProcessingStrategy


class Transparent(ProcessingStrategy):
    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        if len(stream_inputs) != len(stream_outputs):
            raise Exception(
                'it was expected that the number of output stream'
                ' channels is the same as number of input stream channels')

        if len(event_inputs) != len(event_outputs):
            raise Exception(
                'it was expected that the number of output event channels is the same as number of input event channels')

        for i, stream_input in enumerate(stream_inputs):
            np.copyto(stream_outputs[i], stream_input)

        for i, event_input in enumerate(event_inputs):
            for event in event_input:
                event_outputs[i].append(event)

        pass



from copy import copy, deepcopy
from dataclasses import dataclass
from math import ceil
from typing import List, Optional, Dict, Union, cast

import numpy as np

from py_headless_daw.dsp_utils.drum_synth.drum_synth_generator import DrumSynthGenerator
from py_headless_daw.project.plugins.drum_synth_plugin import DrumSynthPlugin
from py_headless_daw.schema.dto.time_interval import TimeInterval
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.midi_event import MidiEvent
from py_headless_daw.schema.events.parameter_value_event import ParameterValueEvent
from py_headless_daw.schema.processing_strategy import ProcessingStrategy

# import matplotlib.pyplot as plt


@dataclass
class Hit:
    data: np.ndarray
    start_sample_in_hit: int
    start_sample_in_buffer: int
    sample_length: int


class DrumSynthStrategy(ProcessingStrategy):

    def __init__(self, plugin: DrumSynthPlugin):

        self._cached_sound: Optional[np.ndarray] = None
        self._unfinished_hits: List[Hit] = []

        # we will use this one as something that holds parameters,
        # tracks parameter changes, and generates generators
        self._plugin: DrumSynthPlugin = deepcopy(plugin)

        super().__init__()

    def render(self, interval: TimeInterval, stream_inputs: List[np.ndarray], stream_outputs: List[np.ndarray],
               event_inputs: List[List[Event]], event_outputs: List[List[Event]]):
        all_events = self._flatten_event_inputs(event_inputs)
        all_events = sorted(all_events, key=lambda event: event.sample_position)

        parameter_value_events = [e for e in all_events if
                                  isinstance(e, ParameterValueEvent)]  # type: List[ParameterValueEvent]

        midi_note_on_events = [e for e in all_events if
                               isinstance(e, MidiEvent) and e.is_note_on()]  # type: List[MidiEvent]

        for event in parameter_value_events:
            # something has changed, so we conservatively
            # invalidate the cache
            #
            # importantly, though, existing hits (i.e. unfinished ones)
            # will still reference the old cache and, therefore, their
            # rendering will be finalized appropriately, after which their
            # remembered cache values will be garbage collected.
            self._invalidate_cache()
            self._plugin.set_parameter_value(event.parameter_id, event.value)

        if self._cached_sound is None:  # the cache was not initialized or has been invalidated
            if self.unit is None:
                raise ValueError('unit is None, which was not expected (error: 0faca3c9)')
            self._regenerate_cache(self.unit.host.sample_rate)

        new_hits = self._convert_note_events_to_new_hits(midi_note_on_events)

        all_hits = [*new_hits, *self._unfinished_hits]

        self._zero_outputs(stream_outputs)

        self._unfinished_hits = self._apply_hits_to_outputs(stream_outputs, all_hits)
        pass

    def _apply_hits_to_outputs(self, stream_outputs: List[np.ndarray], hits: List[Hit]) -> List[Hit]:
        """
        renders all hits to outputs, returns a list hits to be passed to the next iteration

        :param stream_outputs: List of buffers to apply hits to
        :param hits: all hits to apply to current buffers
        :returns: a list of hits to be dealt with on the next iteration
        :rtype: List[Hit]
        """

        hits_for_next_iteration: List[Hit] = []

        for hit in hits:
            hit_for_next: Optional[Hit] = None
            for stream_output in stream_outputs:
                hit_for_next_from_this_channel = self._apply_one_hit_to_one_output(hit, stream_output)
                if hit_for_next is None:
                    # we take the first reported hit for next (presumably from the
                    # first channel)
                    hit_for_next = hit_for_next_from_this_channel

            if hit_for_next is not None:
                hits_for_next_iteration.append(hit_for_next)

        return hits_for_next_iteration

    def _apply_one_hit_to_one_output(self, hit: Hit, stream_output: np.ndarray) -> Optional[Hit]:

        buffer_length = len(stream_output)
        remaining_samples_in_hit = hit.sample_length - hit.start_sample_in_hit
        patch_start_in_output = hit.start_sample_in_buffer
        patch_end_in_output = min(buffer_length, patch_start_in_output + remaining_samples_in_hit)

        patch_length = patch_end_in_output - patch_start_in_output

        patch_start_in_hit = hit.start_sample_in_hit
        patch_end_in_hit = patch_start_in_hit + patch_length

        patched = stream_output[patch_start_in_output:patch_end_in_output]
        patch = hit.data[patch_start_in_hit:patch_end_in_hit]

        np.add(patch, patched, out=patched)

        if patch_end_in_output == buffer_length:
            res = copy(hit)
            res.start_sample_in_hit += patch_length
            res.start_sample_in_buffer = 0
            return res
        else:
            return None

    def _initialize_cache(self):
        # synthesize data to cached_sound
        # detect how long the actual data is until it reliable drops to zero
        self._cached_sound_length_samples = 44100 * 2
        pass

    def _apply_parameter_changes(self, values: Dict[str, Union[str, float]]):
        if len(values):
            self._cached_sound = None

    def _invalidate_cache(self):
        self._cached_sound = None
        self._cached_sound_length_samples = 0

    def _regenerate_cache(self, sample_rate: int):
        generator = DrumSynthGenerator(self._plugin.generate_generator_config())
        needed_length_samples = ceil(generator.get_length_of_full_hit_seconds() * sample_rate)
        self._cached_sound = np.zeros(shape=(needed_length_samples,), dtype=np.float32)
        generator.render_to_buffer(
            output_buffer=self._cached_sound,
            sample_rate=sample_rate,
            start_sample=0)

        # plt.plot(self._cached_sound)
        # plt.show()

    def _convert_note_events_to_new_hits(self, events: List[MidiEvent]) -> List[Hit]:
        res: List[Hit] = []
        cached_sound = cast(np.ndarray, self._cached_sound)
        for event in events:
            hit = Hit(data=self._cached_sound,
                      sample_length=cached_sound.shape[0],
                      start_sample_in_buffer=event.sample_position,
                      start_sample_in_hit=0)
            res.append(hit)

        return res

    def _zero_outputs(self, stream_outputs: List[np.ndarray]):
        for o in stream_outputs:
            o.fill(0)



from typing import List, Dict, Optional, Callable, cast

from py_headless_daw.compiler.internal_plugin_processing_strategy_factory import InternalPluginProcessingStrategyFactory
from py_headless_daw.processing.event.midi_track_strategy import MidiTrackStrategy
from py_headless_daw.processing.event.value_provider_based_event_emitter import ValueProviderBasedEventEmitter
from py_headless_daw.processing.hybrid.vst_plugin import VstPlugin as VstPluginProcessingStrategy
from py_headless_daw.project.audio_track import AudioTrack
from py_headless_daw.project.midi_track import MidiTrack
from py_headless_daw.project.parameter import Parameter
from py_headless_daw.project.plugins.internal_plugin import InternalPlugin as InternalProjectPlugin
from py_headless_daw.project.plugins.plugin import Plugin
from py_headless_daw.project.plugins.vst_plugin import VstPlugin as VstProjectPlugin
from py_headless_daw.project.project import Project
from py_headless_daw.project.track import Track
from py_headless_daw.schema.chain import Chain
from py_headless_daw.schema.events.event import Event
from py_headless_daw.schema.events.parameter_value_event import ParameterValueEvent
from py_headless_daw.schema.host import Host
from py_headless_daw.schema.unit import Unit
from py_headless_daw.schema.wiring import StreamNode, Connector


class ProjectCompiler:

    def __init__(self, internal_plugin_processing_strategy_factory: InternalPluginProcessingStrategyFactory):
        self._internal_plugin_processing_strategy_factory: InternalPluginProcessingStrategyFactory = \
            internal_plugin_processing_strategy_factory

    def compile(self, project: Project) -> List[StreamNode]:
        """
        Given a project, returns a list of stream nodes (as many nodes as many channels
        there are - e.g. 2 for stereo.

        Then the user is expected to iteratively render data from those nodes into buffers.

        :param project: Project
        :return: List[StreamNode]
        """

        # this one is to keep track of compiler tracks to properly handle cycles
        compiled_tracks: Dict[Track, Unit] = {}

        host = Host()

        master_unit = self._compile_internal(host, project, project.master_track, compiled_tracks)
        return master_unit.output_stream_nodes

    def _compile_internal(self, host: Host, project: Project, track: Track, compiled_tracks: Dict[Track, Unit]) -> Unit:
        """
        compiles the "track itself" and then descends to its inputs
        """

        this_track_unit: Unit = self._compile_track_itself(host, project, track, compiled_tracks)

        for input_track in track.inputs:
            input_unit = self._compile_internal(host, project, input_track, compiled_tracks)
            self._connect_units(input_unit, this_track_unit)

        return this_track_unit

    def _compile_track_itself(self, host: Host, project: Project, track: Track,
                              compiled_tracks: Dict[Track, Unit]) -> Unit:
        """
        non-recursively compiles a track

        :param track: Track
        :param compiled_tracks: Dict[Track,Unit]
        :return:
        """
        if track in compiled_tracks:
            return compiled_tracks[track]
        if isinstance(track, MidiTrack):
            res = self._compile_midi_track_itself(host, project, track)
        elif isinstance(track, AudioTrack):
            res = self._compile_audio_track_itself(host, project, cast(AudioTrack, track))
        else:
            raise Exception(f'don\'t know how to compile this type of track: {str(type(track))}')

        compiled_tracks[track] = res
        return res

    @classmethod
    def _connect_units(cls, source: Unit, destination: Unit):
        for i, source_stream_node in enumerate(source.output_stream_nodes):
            output_stream_node = destination.input_stream_nodes[i]
            if not Connector.connected(source_stream_node, output_stream_node):
                Connector(source_stream_node, output_stream_node)

        for i, source_event_node in enumerate(source.output_event_nodes):
            output_event_node = destination.input_event_nodes[i]
            if not Connector.connected(source_event_node, output_event_node):
                Connector(source_event_node, output_event_node)

    # noinspection PyUnusedLocal
    @staticmethod
    def _compile_midi_track_itself(host: Host, project: Project, track: MidiTrack) -> Unit:
        strategy = MidiTrackStrategy(track)
        unit = Unit(0, 0, 0, 1, host, strategy)
        unit.name = "midi track"
        return unit

    def _compile_audio_track_itself(self, host: Host, project: Project, track: AudioTrack) -> Chain:

        previous_unit: Optional[Unit] = None
        first_unit: Optional[Unit] = None
        last_unit: Optional[Unit] = None

        for number, plugin in enumerate(track.plugins):
            last_unit = self._create_audio_plugin_unit(host, project, plugin, track)
            if 0 == number:
                first_unit = last_unit
            if previous_unit is not None:
                self._connect_units(previous_unit, last_unit)
            previous_unit = last_unit

        if first_unit is None or last_unit is None:
            raise Exception('first_unit and/or last_unit are not defined')

        res = Chain(first_unit, last_unit)
        res.name = track.name
        return res

    def _create_audio_plugin_unit(self, host: Host, project: Project, plugin: Plugin, track: Track) -> Unit:

        main_unit: Unit

        if isinstance(plugin, VstProjectPlugin):
            main_unit = self._create_vst_audio_plugin_unit(host, project, plugin, track)
        elif isinstance(plugin, InternalProjectPlugin):
            main_unit = self._create_internal_plugin_unit(host, project, plugin, track)
        else:
            raise Exception('do not know how to treat project plugins of class ' + type(plugin).__name__)

        for parameter in plugin.parameters:
            if parameter.value_provider is not None:
                def parameter_value_transformer(value: float, sample_position: int) -> ParameterValueEvent:
                    # the "parameter" is enclosed from the outer scope
                    return ParameterValueEvent(sample_position, parameter.name, value)

                value_event_emitter_unit: Unit = self._create_unit_for_parameter_value_emission(
                    host,
                    parameter,
                    parameter_value_transformer)

                Connector(value_event_emitter_unit.output_event_nodes[0], main_unit.input_event_nodes[0])

        return main_unit

    @classmethod
    def _create_vst_audio_plugin_unit(cls, host: Host, project: Project, plugin: VstProjectPlugin,
                                      track: Track) -> Unit:
        if plugin.is_synth:
            num_input_event_channels = 1
            num_input_stream_channels = 0
            num_output_event_channels = 0
            num_output_stream_channels = project.num_audio_channels
        else:
            num_input_event_channels = 1  # for envelopes in the future
            num_input_stream_channels = project.num_audio_channels
            num_output_event_channels = 0
            num_output_stream_channels = project.num_audio_channels

        unit = Unit(num_input_stream_channels, num_input_event_channels, num_output_stream_channels,
                    num_output_event_channels, host)

        strategy = cls._create_vst_plugin_processing_strategy(plugin.path_to_shared_library, unit, plugin)
        unit.set_processing_strategy(strategy)
        unit.name = track.name + " - " + plugin.name
        return unit

    def _create_internal_plugin_unit(self, host: Host, project: Project, plugin: InternalProjectPlugin,
                                     track: Track) -> Unit:
        strategy = self._internal_plugin_processing_strategy_factory.produce(plugin, track)

        num_input_event_channels = 1
        num_input_stream_channels = project.num_audio_channels
        num_output_event_channels = 1
        num_output_stream_channels = project.num_audio_channels

        unit = Unit(num_input_stream_channels, num_input_event_channels, num_output_stream_channels,
                    num_output_event_channels, host, strategy)

        unit.name = track.name + ' - ' + type(strategy).__name__
        return unit

    @classmethod
    def _create_unit_for_parameter_value_emission(cls, host: Host, parameter: Parameter,
                                                  transformer_function: Callable[
                                                      [float, int], Event]) -> Unit:
        if parameter.value_provider is None:
            raise Exception('value_provider is None')

        strategy = ValueProviderBasedEventEmitter(parameter.value_provider, transformer_function)
        res = Unit(0, 0, 0, 1, host, strategy)

        res.name = "parameter emitter for %s" % (parameter.name,)

        return res

    @classmethod
    def _create_parameter_value_transformer_function(cls, parameter, plugin) -> Callable[[float], Event]:
        pass

    @classmethod
    def _create_vst_plugin_processing_strategy(cls, path_to_shared_lib: str, unit: Unit, plugin: VstProjectPlugin):

        path_to_shared_lib_as_bytes: bytes = path_to_shared_lib.encode('utf-8')
        strategy = VstPluginProcessingStrategy(path_to_shared_lib_as_bytes, unit)

        for parameter in plugin.parameters:
            strategy.set_parameter_value(parameter.name, parameter.value)

        return strategy
